// #*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#* DtMx4kPlayerDemo.cpp *#*#*#*#*#*# (C) 2014-2019 DekTec
//
// This example application demonstrates a matrix configuration for a 4k (UHD) player. The
// application uses two callbacks: one to generate the 4k image (red-white-blue flag) with
// orange floating rectangles overlayed and one to generate the audio (2x stereo, with
// a fixed tone).
//
// NOTE: 
//  - This application uses all 4 ports of a DTA-2174 to generate SMPTE-425-5 compliant
//    12G (4x3G) 2160P60 output
//  - For a single output (SD/HD/3G) player example please refer to the DtMxPlayerDemo app
//

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Include files -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
#include <cstdio>
#include <cstring>
#include <cstdlib>
#include <vector>
#ifdef _WIN32
    // Windows specific includes
    #include <windows.h>
    #include <conio.h>
#else
    // Linux specific includes
    #include <unistd.h>
    #include <limits.h>
    #include <termios.h>
    #include <sys/select.h>
#endif
#include "Mx4kPlayerDemo.h"

// All code resides in MxExamples namespace
using namespace MxExamples;

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Application Version -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
#define APP_NAME                "DtMx4kPlayerDemo"
#define APP_VERSION_MAJOR       1
#define APP_VERSION_MINOR       5
#define APP_VERSION_BUGFIX      0

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Application configuration -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
// Card to attach to
static const int  CARD_TYPE = 2174;
// Number of card to attach to
static const int  CARD_NO = 0;
// Row indices
static const int  OUT_ROW=0;
// Number of floating rectagles
static const int  NUM_RECTANGLES = 20;
// First physical port in the port range (1 port for 12G or 4 ports for QuadLink 3G)
static const int  FIRST_PORT = 1;
// Number of stereo channel pairs on which to generate audio
static const int  NUM_CHANNEL_PAIRS = 2;
// Video standard. Only use 4k standards/formats for quadlink.
static const int  VIDSTD = DTAPI_VIDSTD_2160P50;

// Link standard, will be set based upon quadlink capability.
static int  LinkStd;
// Last physical port in the port range, will be set based upon quadlink capability.
static int  LastPort;

// Comment out this define to remove profiling information
#define  MX_PRINT_PROFILING_INFO

//=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+ Main loop +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- main -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
int main(int  argc, char*  argv[])
{

    DTAPI_RESULT  dr;
    DtDevice  TheCard;
    Mx4kPlayerDemo  PlayerDemo;

#ifndef _WIN32
    ChangeTerminalMode(1);  // Set terminal mode, so that we can use _kbhit()
#endif

    try
    {
        //-.-.-.-.-.-.-.-.-.-.-.-.-.- Display welcome message -.-.-.-.-.-.-.-.-.-.-.-.-.-.

        printf("%s V%d.%d.%d (c) 2014-2019 DekTec Digital Video B.V.\n\n", APP_NAME,
                                APP_VERSION_MAJOR, APP_VERSION_MINOR, APP_VERSION_BUGFIX);

        //.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Attach to the card -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.

        dr = TheCard.AttachToType(CARD_TYPE, CARD_NO);
        if (dr != DTAPI_OK)
            MX_THROW_EXC(dr, "Failed to attach to DTA-%d", CARD_TYPE);

        //.-.-.-.-.-.-.-.-.-.-.-.-.- Detect Single vs Quadlink .-.-.-.-.-.-.-.-.-.-.-.-.-.

        if ((TheCard.m_pHwf[FIRST_PORT-1].m_Flags&DTAPI_CAP_QUADLINK) != 0)
        {
            // Choose a 3G quadlink 4k link standard and set port range.
            // Note that only 4K Video Standards should be chosen with quadlink.
            LinkStd = DTAPI_VIDLNK_4K_SMPTE425;
            LastPort = FIRST_PORT + 3;
        }
        else
        {
            // Choose a 12G single link 4k link standard and set port range to just one
            // port.
            LinkStd = DTAPI_VIDLNK_4K_SMPTE2082;
            LastPort = FIRST_PORT;
        }

        //.-.-.-.-.-.-.-.-.-.-.-.-.- Check Matrix2 Capability -.-.-.-.-.-.-.-.-.-.-.-.-.-.

        // Check the ports for CAP_MATRIX2. The DtMxXxx API is only supported by ports 
        // with DTAPI_CAP_MATRIX2 capability
        for (int  i=(FIRST_PORT-1); i<=(LastPort-1); i++)
        {
            if ((TheCard.m_pHwf[i].m_Flags&DTAPI_CAP_MATRIX2) == 0)
            {
                MX_THROW_EXC_NO_ERR("Port %d does not have the MATRIX2 capability", 
                                                                TheCard.m_pHwf[i].m_Port);
            }
        }
                       
        //.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Create the matrix -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

        // Init the matrix
        if (!PlayerDemo.Init(TheCard))
            MX_THROW_EXC_NO_ERR("Failed to init demo");
        // Start the matrix
        if (!PlayerDemo.Start())
            MX_THROW_EXC_NO_ERR("Failed to start demo");

        //-.-.-.-.-.-.-.-.-.- Process loop, wait for user to press key -.-.-.-.-.-.-.-.-.-
        
        printf("\nPress any key to exit\n\n");
        while(!_kbhit()) { Sleep(10); }
        _getch();   // Read the keystroke from the console

        PlayerDemo.Stop(); // Stop the matrix

        // Print profling information
#ifdef MX_PRINT_PROFILING_INFO
        PlayerDemo.PrintProfilingInfo();
        printf("\nPress any key to exit\n");
        // Wait for a keystroke before exiting. To allow user time to take in the 
        // profiling information
        _getch();
#endif
    }
    catch(MxDemoExc  e)
    {
        puts(e.ErrorStr());
    }
    catch(...) 
    {
        puts("Encountered an unhandled exception!!!");
    }

#ifndef _WIN32
    ChangeTerminalMode(0);  // Restore terminal mode
#endif
    return 0;
}

// All code resides in MxExamples namespace
namespace MxExamples
{

//=+=+=+=+=+=+=+=+=+=+=+=+=+=+ Mx4kPlayerDemo implementation +=+=+=+=+=+=+=+=+=+=+=+=+=+=+

//.-.-.-.-.-.-.-.-.-.-.-.-.-.- Mx4kPlayerDemo::Mx4kPlayerDemo -.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
Mx4kPlayerDemo::Mx4kPlayerDemo()
{
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.- Mx4kPlayerDemo::~Mx4kPlayerDemo -.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
Mx4kPlayerDemo::~Mx4kPlayerDemo()
{
    Stop(); // Call stop, just in case we did not stop it ourselves
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Mx4kPlayerDemo::Init -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
bool  Mx4kPlayerDemo::Init(DtDevice&  TheCard)
{
    DTAPI_RESULT  dr;

    //-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Prep card for matrix -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

    // The card must have been attached
    if (!TheCard.IsAttached())
        return false;
    if (!PrepCard(TheCard))
        return false;

    //+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+ Configure the 4k row +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+

    //.-.-.-.-.-.-.-.-.-.-.-.-.-.- Step 1: Create an 4k port -.-.-.-.-.-.-.-.-.-.-.-.-.-.-
    // Add one port or associate four ports to create one logical 4k port (2160pXX)
    DtMxPort  Port4k(VIDSTD, LinkStd);
    for (int  i=(FIRST_PORT-1); i<=(LastPort-1); i++)
    {
        dr = Port4k.AddPhysicalPort(&TheCard, i+1);
        if (dr != DTAPI_OK)
            MX_THROW_EXC(dr, "Failed to add port %d to Port4k", (i+1));
    }

    //-.-.-.-.-.-.-.-.-.-.-.-.- Step 2: attach 4k port to a row -.-.-.-.-.-.-.-.-.-.-.-.-.
    // Set the logical 4k port as the row output
    dr = m_TheMatrix.AttachRowToOutput(OUT_ROW, Port4k);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed to attach row %d to the logical 4k port", OUT_ROW);

    printf("  Video standard = %s\n", ::DtapiVidStd2Str(VIDSTD));
    printf("  Link standard = %s\n", ::DtapiLinkStd2Str(LinkStd));

    //.-.-.-.-.-.-.-.-.-.-.-.-.- Step 3: set row configuration -.-.-.-.-.-.-.-.-.-.-.-.-.-

    DtMxRowConfig  RowConfig;
    RowConfig.m_Enable = true;
    RowConfig.m_RowSize = 1;       // Keep a history of one frame

    // Enable the video
    RowConfig.m_VideoEnable = true;
    RowConfig.m_Video.m_PixelFormat = DT_PXFMT_YUV422P_8B;  // Use a planar format (8-bit)
    RowConfig.m_Video.m_StartLine1 = RowConfig.m_Video.m_StartLine2 = 1;
    RowConfig.m_Video.m_NumLines1 = RowConfig.m_Video.m_NumLines2 = -1;
    RowConfig.m_Video.m_Scaling = DTAPI_SCALING_OFF;   // No scaling for this row
    RowConfig.m_Video.m_LineAlignment = 8; // Want an 8-byte alignment

    // Enable audio
    RowConfig.m_AudioEnable = true;
    // Add channel pairs (stereo)
    printf("  Number of stereo channel pairs being generated = %d\n", NUM_CHANNEL_PAIRS);
    for (int  i=0; i< NUM_CHANNEL_PAIRS*2; i++)
    {
        DtMxAudioConfig  AudioConfig;
        AudioConfig.m_Index = i;
        AudioConfig.m_DeEmbed = false;  // Nothing to de-embed for an output only row
        AudioConfig.m_OutputMode = DT_OUTPUT_MODE_ADD;  // We want to add audio samples
        AudioConfig.m_Format = DT_AUDIO_SAMPLE_PCM;     // We will add PCM samples

        // Add to config list
        RowConfig.m_Audio.push_back(AudioConfig);
    }

    // Apply the row configuration
    dr = m_TheMatrix.SetRowConfig(OUT_ROW, RowConfig);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed set row %d config", OUT_ROW);

    //-.-.-.-.-.-.-.-.-.-.-.-.-.- Step 4: Final configuration -.-.-.-.-.-.-.-.-.-.-.-.-.-.

    // Determine the default recommended end-to-end delay for this configuration
    int  DefEnd2EndDelay = 0;
    double CbFrames=0.0;    // Time available for the call-back method
    dr = m_TheMatrix.GetDefEndToEndDelay(DefEnd2EndDelay, CbFrames);
    printf("  Default end-to-end delay = %d Frames\n", DefEnd2EndDelay);
    printf("  Time-for-Callback = %.1f Frame Period\n", CbFrames);
    // Apply the default end-to-end delay
    dr = m_TheMatrix.SetEndToEndDelay(DefEnd2EndDelay);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed set the end-to-end delay (delay=%d)", DefEnd2EndDelay);

    // Register our callbacks
    m_ContextAudio.m_pPlayer = this; 
    m_ContextAudio.m_Type = CallbackContext::GEN_AUDIO;
    dr = m_TheMatrix.AddMatrixCbFunc(Mx4kPlayerDemo::OnNewFrame, &m_ContextAudio);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed register audio callback function");
    
    m_ContextVideo.m_pPlayer = this; 
    m_ContextVideo.m_Type = CallbackContext::GEN_VIDEO;
    dr = m_TheMatrix.AddMatrixCbFunc(Mx4kPlayerDemo::OnNewFrame, &m_ContextVideo);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed register video callback function");

    //-.-.-.-.-.-.-.-.-.-.-.-.-.- Init our floating rectangles -.-.-.-.-.-.-.-.-.-.-.-.-.-

    DtVidStdInfo  VidInfo;
    dr = ::DtapiGetVidStdInfo(VIDSTD, LinkStd, VidInfo);

    // Set bounding rectangle to size of image
    m_BoundingRect = MxRectangle(0,0, VidInfo.m_VidWidth, VidInfo.m_VidHeight,
                                                                            Color(0,0,0));
    
    // Seed random generator
    srand(NUM_RECTANGLES);
    MxPoint  Centre = m_BoundingRect.Centre();
    for (int  i=0; i<NUM_RECTANGLES; i++)
    {
        // Randomise size of the rectangle
        // NOTE: the x2 is because we want all our sizes to be multiples of 2
        int  Size = (((rand() % 100)*2) + 50) * 2;

        // Clip size to a maximum of 1/3th of the video height
        if (Size > (m_BoundingRect.Height() / 3))
            Size = (m_BoundingRect.Height() / 6) * 2;

        // Create the a rectangle with the specified size
        MxRectangle  Rect = MxRectangle(0, 0, Size, Size, COL_ORANGE);

        // Start at the centre
        MxPoint  P = Centre;
        P.Offset(-Size/2, -Size/2);
        Rect.MoveTo(P);

        // Randomise movement vector
        int  SpeedX = (rand()%6 + 1) * 2;
        int  SpeedY = (rand()%7 + 1) * 2;
        // Randomise direction
        switch (rand()%4)
        {
        // Move left and upwards
        case 0:  Rect.m_Vector = MxPoint(-SpeedX, -SpeedY); break;
        // Move left and downwards
        case 1:  Rect.m_Vector = MxPoint(-SpeedX, SpeedY); break;
        // Move right and upwards
        case 2:  Rect.m_Vector = MxPoint(SpeedX, -SpeedY); break;
        // Move right and downwards
        case 3:  Rect.m_Vector = MxPoint(SpeedX, SpeedY); break;
        }

        // Add MxRectangle to list
        m_Rectangles.push_back(Rect);
    }

    // .-.-.-.-.-.-.-.-.-.-.-.-.-.- Init the tone generators -.-.-.-.-.-.-.-.-.-.-.-.-.-.-
    // Create a tone generator with distinct frequency and volume for each channel.
    
    const int  NumAudioChannels = NUM_CHANNEL_PAIRS*2;
    const double  MaxAttenuation = .4;
    const double  IncrementalAttenuation = MaxAttenuation / NumAudioChannels;

    m_ToneGenerators.clear();
    for (int i=0; i < NumAudioChannels; ++i)
    {
        int  ToneFreq = 1000 + i*250;
        double  Attenuation = IncrementalAttenuation * (i+1);

        m_ToneGenerators.push_back(MxToneGenerator(ToneFreq, Attenuation));
    }

    return true;
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Mx4kPlayerDemo::PrepCard -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
bool  Mx4kPlayerDemo::PrepCard(DtDevice&  TheCard)
{
    DTAPI_RESULT  dr;
           
    //-.-.-.-.-.-.-.-.-.-.-.-.-.- Configure port io-direction -.-.-.-.-.-.-.-.-.-.-.-.-.-.

    DtIoConfig  Cfg;
    std::vector<DtIoConfig>  IoConfigs;

    for (int  i=(FIRST_PORT-1); i<=(LastPort-1); i++)
    {
        Cfg.m_Port = (i+1);
        Cfg.m_Group = DTAPI_IOCONFIG_IODIR;
        Cfg.m_Value = DTAPI_IOCONFIG_OUTPUT;
        Cfg.m_SubValue = DTAPI_IOCONFIG_OUTPUT;
        Cfg.m_ParXtra[0] = Cfg.m_ParXtra[1] = -1;
        // Add to list
        IoConfigs.push_back(Cfg);
    }
    dr = TheCard.SetIoConfig(IoConfigs.data(), (int)IoConfigs.size());
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed to apply IO-DIR for ports");

    //-.-.-.-.-.-.-.-.-.-.-.-.-.- Configure ports io-standard -.-.-.-.-.-.-.-.-.-.-.-.-.-.
    // Configure IO-STD for all 4 ports 

    int  IoStdValue=-1, IoStdSubValue=-1;
    dr = ::DtapiVidStd2IoStd(VIDSTD, LinkStd, IoStdValue, IoStdSubValue);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed to deduce IO-STD from video and link standard");

    IoConfigs.clear();
    for (int  i=(FIRST_PORT-1); i<=(LastPort-1); i++)
    {
        Cfg.m_Port = i+1;
        Cfg.m_Group = DTAPI_IOCONFIG_IOSTD;
        Cfg.m_Value = IoStdValue;
        Cfg.m_SubValue = IoStdSubValue;
        Cfg.m_ParXtra[0] = Cfg.m_ParXtra[1] = -1;
    }
    dr = TheCard.SetIoConfig(IoConfigs.data(), (int)IoConfigs.size());
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed to apply IO-STD for ports");

    //.-.-.-.-.-.-.-.-.-.-.-.- Configure the internal GENREF port -.-.-.-.-.-.-.-.-.-.-.-.
    // NOTE: setup the internal GENREF port to generate a frame timing that matches the 
    // output video standard, so that we ensure all ports are aligned

    IoConfigs.clear();
    DtHwFuncDesc*  pHwf = TheCard.m_pHwf;
    for (int  i=0; i<TheCard.m_pHwf->m_DvcDesc.m_NumPorts; i++, pHwf++)
    {
        // Check if the port has GENNREF capability
        if ((pHwf->m_Flags & DTAPI_CAP_GENREF) == 0)
            continue;   // No GENREF cap

        Cfg.m_Port = pHwf->m_Port;
        Cfg.m_Group = DTAPI_IOCONFIG_GENREF;

        // Check if this is the internal GENREF (has DTAPI_CAP_VIRTUAL)?
        if ((pHwf->m_Flags & DTAPI_CAP_VIRTUAL) != 0)
            Cfg.m_Value = DTAPI_IOCONFIG_TRUE;   // The internal => enable as GENREF
        else
            Cfg.m_Value = DTAPI_IOCONFIG_FALSE;  // Not the internal => disable as GENREF
        Cfg.m_SubValue = -1;
        Cfg.m_ParXtra[0] = Cfg.m_ParXtra[1] = -1;
        // Add to list
        IoConfigs.push_back(Cfg);

        // Set the IO-standard for 'THE' GENREF port
        if (Cfg.m_Value != DTAPI_IOCONFIG_TRUE)
            continue;
        Cfg.m_Group = DTAPI_IOCONFIG_IOSTD;
        // Set a standard with matching frame rate
        DtVidStdInfo  VidInfo;
        dr = ::DtapiGetVidStdInfo(VIDSTD, LinkStd, VidInfo);
        Cfg.m_Value=DTAPI_IOCONFIG_HDSDI; 
        int  Fps = (int)VidInfo.m_Fps;
        switch (VIDSTD)
        {
        case DTAPI_VIDSTD_2160P50B:    Fps = 25; break;
        case DTAPI_VIDSTD_2160P59_94B: Fps = 29; break;
        case DTAPI_VIDSTD_2160P60B:    Fps = 30; break;
        }
        switch (Fps)
        {
        case 23: Cfg.m_SubValue = DTAPI_IOCONFIG_720P23_98; break;
        case 24: Cfg.m_SubValue = DTAPI_IOCONFIG_720P24; break;
        case 25: Cfg.m_SubValue = DTAPI_IOCONFIG_720P25; break;
        case 29: Cfg.m_SubValue = DTAPI_IOCONFIG_720P29_97; break;
        case 30: Cfg.m_SubValue = DTAPI_IOCONFIG_720P30; break;
        case 50: Cfg.m_SubValue = DTAPI_IOCONFIG_720P50; break;
        case 59: Cfg.m_SubValue = DTAPI_IOCONFIG_720P59_94; break;
        case 60: Cfg.m_SubValue = DTAPI_IOCONFIG_720P60; break;
        }
        Cfg.m_ParXtra[0] = Cfg.m_ParXtra[1] = -1;
        // Add to list
        IoConfigs.push_back(Cfg);
    }
    dr = TheCard.SetIoConfig(IoConfigs.data(), (int)IoConfigs.size());
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed to apply GENREF IO-config");

    // Wait for until we are locked to the genlock-reference-source
    {
        const int  WAIT_TIME_MS=250, MAX_WAIT_TIME_MS=10000, MIN_WAIT_TIME_MS=500;
        const int  MIN_LOOP_CNT=(MIN_WAIT_TIME_MS/WAIT_TIME_MS);
        const int  MAX_LOOP_CNT=(MAX_WAIT_TIME_MS/WAIT_TIME_MS);
        int  LoopCnt=0, State=DTAPI_GENL_NO_REF;
        do
        {
            const char  ActivityIndicator[4] = {'|', '/', '-', '\\'};
            const char*  StateStr;
            if (State == DTAPI_GENL_LOCKED)         StateStr = "LOCKED";
            else if (State == DTAPI_GENL_NO_REF)    StateStr = "NO REF";
            else if (State == DTAPI_GENL_LOCKING)   StateStr = "LOCKING";
            else                                    StateStr = "NO LOCK";

            printf("%c Waiting for GenRef lock (current state=%s)  \r", 
                                                  ActivityIndicator[LoopCnt%4], StateStr);
            
            Sleep(WAIT_TIME_MS);
            
            // Get lock state
            dr = TheCard.GetGenlockState(State);
            if (dr != DTAPI_OK)
                MX_THROW_EXC(dr, "GetGenlockState failed");

            if (_kbhit())
            {
                _getch();   // read keystroke from console
                return false;
            }
            LoopCnt++;
        }
        while((State!=DTAPI_GENL_LOCKED || LoopCnt<MIN_LOOP_CNT) && LoopCnt<MAX_LOOP_CNT);

        printf("\n");
        if (State != DTAPI_GENL_LOCKED)
            printf("WARNING: did not lock to our genlock-reference\n");
    }
    
    return (dr == DTAPI_OK);
}

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Mx4kPlayerDemo::OnNewFrame -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
void  Mx4kPlayerDemo::OnNewFrame(DtMxData* pData, void* pOpaque)
{
    const CallbackContext*  pContext = ((const CallbackContext*)pOpaque);
    if (pContext->m_Type == CallbackContext::GEN_VIDEO)
        pContext->m_pPlayer->OnGenerateVideo(pData);
    else if (pContext->m_Type == CallbackContext::GEN_AUDIO)
        pContext->m_pPlayer->OnGenerateAudio(pData);
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.- Mx4kPlayerDemo::OnGenerateVideo -.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
void  Mx4kPlayerDemo::OnGenerateVideo(DtMxData* pData)
{
    DtMxRowData&  OurRow = pData->m_Rows[OUT_ROW];

    // Must have a valid frame
    if (OurRow.m_CurFrame->m_Status != DT_FRMSTATUS_OK)
    {
        // Frame is not valid so we have to assume the frame buffers are unusable
        return;
    }

    // Check if any frames were skipped. Indication that we may have a performance issue
    if (pData->m_NumSkippedFrames>0)
        printf("[%lld] #skipped frames = %d\n", pData->m_Frame,
                                                               pData->m_NumSkippedFrames);

    //.-.-.-.-.-.-.-.-.-.-.-.-.-.- Generate the three planes -.-.-.-.-.-.-.-.-.-.-.-.-.-.-
    // We will generate the Dutch flag in the background and overlay a moving orange 
    // rectangle

    //.-.-.-.-.-.-.-.- Step 1 : generate red, white and blue back ground -.-.-.-.-.-.-.-.-

    DtMxVideoBuf&  VideoBuf = OurRow.m_CurFrame->m_Video[0];
    // Get luminance plane
    DtMxVideoPlaneBuf&  PlaneY = VideoBuf.m_Planes[0];
    // Get chrominance plane (Cb=U)
    DtMxVideoPlaneBuf&  PlaneU = VideoBuf.m_Planes[1];
    // Get chrominance plane (Cr=V)
    DtMxVideoPlaneBuf&  PlaneV = VideoBuf.m_Planes[2];

    unsigned char* pY = PlaneY.m_pBuf;
    unsigned char* pU = PlaneU.m_pBuf;
    unsigned char* pV = PlaneV.m_pBuf;
    Color  Col(0,0,0);        
    for (int  i=0; i<VideoBuf.m_Height; i++)
    {
        // First third is the red band
        if (i<(VideoBuf.m_Height/3))
        {
            Col = COL_RED;
        }
        // Second third is the white band
        else if (i<((VideoBuf.m_Height/3)*2))
        {
            Col = COL_WHITE;
        }
        // Third third is the blue band
        else
        {
            Col = COL_BLUE;
        }
        
        // Set luminance plane
        memset(pY, Col.m_Y, VideoBuf.m_Width);
        // Set chrominance plane U
        memset(pU, Col.m_U, VideoBuf.m_Width/2);   // Only half the U-samples
        // Set chrominance plane V
        memset(pV, Col.m_V, VideoBuf.m_Width/2);   // Other half is V-samples

        // Next line
        pY += PlaneY.m_Stride;
        pU += PlaneU.m_Stride;
        pV += PlaneV.m_Stride;
    }

    //.-.-.-.-.-.-.-.-.-.-.-.-.- Step 2: overlay the rectangles -.-.-.-.-.-.-.-.-.-.-.-.-.

    for (int  i=0; i<(int)m_Rectangles.size(); i++)
    {
        MxRectangle&  R = m_Rectangles[i];

        // Move rectangle to new position
        R.Offset(R.m_Vector);

        // Bounce of the left and right edge
        if (R.Left() < m_BoundingRect.Left())
        {
            R.Offset(-R.Left()*2, 0);
            // Invert horizontal direction of movement vector
            R.m_Vector.m_X = -R.m_Vector.m_X;
        }
        else if (R.Right() > m_BoundingRect.Right())
        {
            R.Offset((m_BoundingRect.Right()-R.Right())*2, 0);
            // Invert horizontal direction of movement vector
            R.m_Vector.m_X = -R.m_Vector.m_X;
        }

        // Bounce of the top and bottom edge
        if (R.Top() < m_BoundingRect.Top())
        {
            R.Offset(0, -R.Top()*2);
            // Invert vertical direction of movement vector
            R.m_Vector.m_Y = -R.m_Vector.m_Y;
        }
        else if (R.Bottom() > m_BoundingRect.Bottom())
        {
            R.Offset(0, (m_BoundingRect.Bottom()-R.Bottom())*2);
            // Invert vertical direction of movement vector
            R.m_Vector.m_Y = -R.m_Vector.m_Y;
        }
    
        pY = PlaneY.m_pBuf + (PlaneY.m_Stride * R.Top());
        pU = PlaneU.m_pBuf + (PlaneU.m_Stride * R.Top());
        pV = PlaneV.m_pBuf + (PlaneV.m_Stride * R.Top());
        for (int  j=0; j<R.Height(); j++)
        {
            memset(pY+R.Left(), R.m_Color.m_Y, R.Width());
            memset(pU+R.Left()/2, R.m_Color.m_U, R.Width()/2);
            memset(pV+R.Left()/2, R.m_Color.m_V, R.Width()/2);

            // Next line
            pY += PlaneY.m_Stride;
            pU += PlaneU.m_Stride;
            pV += PlaneV.m_Stride;
        }
    }
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.- Mx4kPlayerDemo::OnGenerateAudio -.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
void  Mx4kPlayerDemo::OnGenerateAudio(DtMxData* pData)
{
    DTAPI_RESULT  dr;

    DtMxRowData&  OurRow = pData->m_Rows[OUT_ROW];

    // Must have a valid frame
    if (OurRow.m_CurFrame->m_Status != DT_FRMSTATUS_OK)
    {
        // Frame is not valid so we have to assume the frame buffers are unusable
        return;
    }

    // Get frame audio data
    DtMxAudioData&  AudioData = OurRow.m_CurFrame->m_Audio;

    int  NumChannels = NUM_CHANNEL_PAIRS*2;

    //.-.-.-.-.-.-.-.-.-.-.-.-.- Create the service description -.-.-.-.-.-.-.-.-.-.-.-.-.
    
    for (int pair = 0; pair < NUM_CHANNEL_PAIRS; ++pair)
    {
        DtMxAudioService&  Service = AudioData.m_Services[pair];
        Service.m_Valid = true;
        Service.m_SampleRate = 48000;      // 48Khz
        Service.m_ContainsData = false;    // We will insert PCM not data
        Service.m_PcmNumBits = 24;         // 24-bit audio samples
        Service.m_ServiceType = DT_AUDIOSERVICE_STEREO;    // Create a stereo pair

        Service.m_Channels.push_back(pair*2);   // Channel n
        Service.m_Channels.push_back(pair*2+1); // Channel n+1
    }
    //-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Add PCM samples -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.

   

    // Generate audio for each stereo service
    for (int  i=0; i<(int)AudioData.m_Services.size(); i++)
    {
        const DtMxAudioService&  Service = AudioData.m_Services[i];
        if (!Service.m_Valid)
            continue;   // Skip invalid services

        // Generate PCM samples for each channel in the service
        for (int  j=0; j<(int)Service.m_Channels.size(); j++)
        {
            int  ChanIndex = Service.m_Channels[j];
            DtMxAudioChannel&  Channel = AudioData.m_Channels[ChanIndex];
            Channel.m_Service = i;  // Set to the correct service
            // Generate as much audio as needed (=m_NumSamplesHint)
            MX_ASSERT(ChanIndex <= (int)m_ToneGenerators.size());
            m_ToneGenerators[ChanIndex].Generate(Channel.m_pBuf, Channel.m_NumSamplesHint);
            Channel.m_NumValidSamples = Channel.m_NumSamplesHint;
        }
    }

    // Init channel status word for all (valid) audio channels
    dr = AudioData.InitChannelStatus();
}

} // namespace MxExamples
