// *#*#*#*#*#*#*#*#*#*#*#*#*#*#*#* DtMxAvRecorderDemo.cpp *#*#*#*#*#* (C) 2015-2019 DekTec
//
// This example application demonstrates a matrix configuration for a AV recorder. The
// application records all audio streams, it finds, to PCM files and records the video
// as a scaled (1/16) yuv file. The application spawns a seperate thread for writing
// to disk. The matrix callback copies the received audio and video data to a buffer
// from which the record thread reads the AV-data and writes the data to disk. The 
// buffer makes sure we can deal with disk write latencies.
// 

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Include files -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
#include <cstdio>
#include <cstring>
#ifdef _WIN32
    // Windows specific includes
    #include <windows.h>
    #include <conio.h>
#else
    // Linux specific includes
    #include <unistd.h>
    #include <limits.h>
    #include <termios.h>
    #include <sys/select.h>
#endif
#include "MxAvRecorderDemo.h"

// All code resides in MxExamples namespace
using namespace MxExamples;

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Application Version -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
#define APP_NAME                "DtMxAvRecorderDemo"
#define APP_VERSION_MAJOR       1
#define APP_VERSION_MINOR       4
#define APP_VERSION_BUGFIX      0

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Application configuration -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

// Default, card to attach to
static const int  DEF_CARD_TYPE = 2174;
// Default, number of card to attach to
static const int  DEF_CARD_NO = 0;
// Default, output port
static const int  DEF_IN_PORT=1;

// Default input video standard
#define DTAPI_VIDSTD_AUTO  -2   // Extra helper for auto detection of video standard
static const int  DEF_IN_VIDSTD = DTAPI_VIDSTD_AUTO;

// Default audio mode (per channel as 32-bit)
static const MxAvRecorderDemo::AudioMode  DEF_AUDIO_MODE = 
                                                    MxAvRecorderDemo::AUDMODE_CHANNEL_32B;

// Row indices
static const int  IN_ROW=0;

// AV stream configuration (max #frames queue, size of AV buffer and number of buffers)
static const int  MAX_NUM_FRAMES = 16;

static const int  VIDEO_MAX_NUM_STREAMS = 1;
static const int  VIDEO_BUF_SIZE = 8*1024*1024;
static const int  VIDEO_NUM_BUFFERS = VIDEO_MAX_NUM_STREAMS * MAX_NUM_FRAMES;

static const int  AUDIO_MAX_NUM_STREAMS = 16;
static const int  AUDIO_BUF_SIZE = 32*1024;
static const int  AUDIO_NUM_BUFFERS = AUDIO_MAX_NUM_STREAMS * MAX_NUM_FRAMES;

// Filename templates
static const char VIDEO_FILENAME_FMT[] = "video_%dx%d.yuv";
static const char AUDIO_FILENAME_FMT1[] = "audio_channel_%d_%dbit.pcm";
static const char AUDIO_FILENAME_FMT2[] = "audio_service_%d_%dx%dbit.pcm";


// Comment out this define to remove profiling information
#define  MX_PRINT_PROFILING_INFO

//=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+ Main loop +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- main -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
int main(int  argc, char*  argv[])
{
    DTAPI_RESULT  dr;
    DtDevice  TheCard;
    MxAvRecorderDemo  R;    // The recorder

#ifndef _WIN32
    ChangeTerminalMode(1);  // Set terminal mode, so that we can use _kbhit()
#endif

    try
    {
        //-.-.-.-.-.-.-.-.-.-.-.-.-.- Display welcome message -.-.-.-.-.-.-.-.-.-.-.-.-.-.

        printf("%s V%d.%d.%d (c) 2015-2019 DekTec Digital Video B.V.\n\n", APP_NAME,
                                APP_VERSION_MAJOR, APP_VERSION_MINOR, APP_VERSION_BUGFIX);

        //.-.-.-.-.-.-.-.-.-.-.-.- Parse the command-line options -.-.-.-.-.-.-.-.-.-.-.-.

        R.ParseCommandLine(argc, argv);        

        //.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Attach to the card -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.

        dr = TheCard.AttachToType(R.CARD_TYPE(), R.CARD_NO());
        if (dr != DTAPI_OK)
            MX_THROW_EXC(dr, "Failed to attach to DTA-%d", R.CARD_TYPE());

        // Check the port for CAP_MATRIX2. The DtMxXxx API is only supported by ports 
        // with DTAPI_CAP_MATRIX2 capability
        if ((TheCard.m_pHwf[(R.IN_PORT()-1)].m_Flags & DTAPI_CAP_MATRIX2) == 0)
            MX_THROW_EXC_NO_ERR("Port %d does not have the MATRIX2 capability", 
                                                                             R.IN_PORT());

        printf("  Using DTA-%d, Port=%d, Standard=%s\n", R.CARD_TYPE(),
                                               R.IN_PORT(), 
                                               R.IN_VIDSTD()==DTAPI_VIDSTD_AUTO ? 
                                               "auto" : ::DtapiVidStd2Str(R.IN_VIDSTD()));
                        
        //.-.-.-.-.-.-.-.-.-.-.-.-.-.- Create the PIP matrix -.-.-.-.-.-.-.-.-.-.-.-.-.-.-

        // Init the matrix
        if (!R.Init(TheCard))
            MX_THROW_EXC_NO_ERR("Failed to init demo");
        // Start the matrix
        if (!R.Start())
            MX_THROW_EXC_NO_ERR("Failed to start demo");

        //-.-.-.-.-.-.-.-.-.- Process loop, wait for user to press key -.-.-.-.-.-.-.-.-.-
        
        printf("\nPress any key to exit\n\n");
        while(!_kbhit()) { Sleep(10); }
        _getch();   // Read the keystroke from the console

        R.Stop(); // Stop the matrix

        // Print profling information
#ifdef MX_PRINT_PROFILING_INFO
        R.PrintProfilingInfo();
        printf("\nPress any key to exit\n");
        // Wait for a keystroke before exiting. To allow user time to take in the 
        // profiling information
        _getch();
#endif
    }
    catch(MxDemoExc  e)
    {
        puts(e.ErrorStr());
    }
    catch(...) 
    {
        puts("Encountered an unhandled exception!!!");
    }

#ifndef _WIN32
    ChangeTerminalMode(0);  // Restore terminal mode
#endif
    return 0;
}

//+=+=+=+=+=+=+=+=+=+=+=+=+=+ MxAvRecorderDemo implementation +=+=+=+=+=+=+=+=+=+=+=+=+=+=

//.-.-.-.-.-.-.-.-.-.-.-.-.- MxAvRecorderDemo::MxAvRecorderDemo -.-.-.-.-.-.-.-.-.-.-.-.-.
//
MxAvRecorderDemo::MxAvRecorderDemo() 
    : m_pRecThread(NULL), m_Exit(false), m_pRecEvent(NULL), m_pRecLock(NULL)
{
    // Create av-stream descriptors
    for (int  i=0; i<VIDEO_MAX_NUM_STREAMS; i++)
        m_VidStreams.push_back(new AvStream(AvStream::VIDEO_STREAM));
    for (int  i=0; i<AUDIO_MAX_NUM_STREAMS; i++)
        m_AudStreams.push_back(new AvStream(AvStream::AUDIO_STREAM, i));

    // Create av-stream-buffers
    for (int  i=0; i<VIDEO_NUM_BUFFERS; i++)
    {
        AvStreamBuf  AvBuf;
        AvBuf.Alloc(VIDEO_BUF_SIZE);
        m_VidBuffers.push_back(AvBuf);
    }
    for (int  i=0; i<AUDIO_NUM_BUFFERS; i++)
    {
        AvStreamBuf  AvBuf;
        AvBuf.Alloc(AUDIO_BUF_SIZE);
        m_AudBuffers.push_back(AvBuf);
    }
}

//-.-.-.-.-.-.-.-.-.-.-.-.- MxAvRecorderDemo::~MxAvRecorderDemo -.-.-.-.-.-.-.-.-.-.-.-.-.
//
MxAvRecorderDemo::~MxAvRecorderDemo()
{
    Stop();

    // Clean-up av-stream descriptors
    for (int  i=0; i<(int)m_VidStreams.size(); i++)
        delete m_VidStreams[i];
    for (int  i=0; i<(int)m_AudStreams.size(); i++)
        delete m_AudStreams[i];

    // Clean-up av-stream-buffers
    AvStreamBufList::iterator  it = m_VidBuffers.begin();
    for ( ; it!=m_VidBuffers.end(); it++)
        it->Free();
    it = m_AudBuffers.begin();
    for ( ; it!=m_AudBuffers.end(); it++)
        it->Free();
}

//.-.-.-.-.-.-.-.-.-.-.-.-.- MxAvRecorderDemo::ParseCommandLine -.-.-.-.-.-.-.-.-.-.-.-.-.
//
void  MxAvRecorderDemo::ParseCommandLine(int argc, char* argv[])
{
    // Table defining video standard command-line options
    static const DtEnumOptPair CMD_VIDSTD_OPTS[] = 
    {
        // Auto detected
        { L"auto",       DTAPI_VIDSTD_AUTO,         L"auto" },

        // 2160: progressive
        { L"2160p60",    DTAPI_VIDSTD_2160P60,      L"2160p/60" },
        { L"2160p59",    DTAPI_VIDSTD_2160P59_94,   L"2160p/59.94" },
        { L"2160p50",    DTAPI_VIDSTD_2160P50,      L"2160p/50" },

        // 1080: Interlaced
        { L"1080i60",    DTAPI_VIDSTD_1080I60,      L"1080i/60" },
        { L"1080i59",    DTAPI_VIDSTD_1080I59_94,   L"1080i/59.94" },
        { L"1080i50",    DTAPI_VIDSTD_1080I50,      L"1080i/50" },
        
        // 1080: progressive
        { L"1080p60",    DTAPI_VIDSTD_1080P60,      L"1080p/60" },
        { L"1080p59",    DTAPI_VIDSTD_1080P59_94,   L"1080p/59.94" },
        { L"1080p50",    DTAPI_VIDSTD_1080P50,      L"1080p/50" },
        { L"1080p30",    DTAPI_VIDSTD_1080P30,      L"1080p/30" },
        { L"1080p29",    DTAPI_VIDSTD_1080P29_97,   L"1080p/29.97" },
        { L"1080p25",    DTAPI_VIDSTD_1080P25,      L"1080p/25" },
        { L"1080p24",    DTAPI_VIDSTD_1080P24,      L"1080p/24" },
        { L"1080p23",    DTAPI_VIDSTD_1080P23_98,   L"1080p/23.98" },

        // 720: progressive
        { L"720p60",     DTAPI_VIDSTD_720P60,       L"720p/60" },
        { L"720p59",     DTAPI_VIDSTD_720P59_94,    L"720p/59.94" },
        { L"720p50",     DTAPI_VIDSTD_720P50,       L"720p/50" },
        { L"720p30",     DTAPI_VIDSTD_720P30,       L"720p/30" },
        { L"720p29",     DTAPI_VIDSTD_720P29_97,    L"720p/29.97" },
        { L"720p25",     DTAPI_VIDSTD_720P25,       L"720p/25" },
        { L"720p24",     DTAPI_VIDSTD_720P24,       L"720p/24" },
        { L"720p23",     DTAPI_VIDSTD_720P23_98,    L"720p/23.98" },

        // SD: NTSC & PAL
        { L"525i59",     DTAPI_VIDSTD_525I59_94,    L"525i/59.94" },
        { L"625i50",     DTAPI_VIDSTD_625I50,       L"625i/50" },

        OPT_PAIR_END,
    };

    // Table defining audio mode command-line options
    static const DtEnumOptPair CMD_AUDMODE_OPTS[] = 
    {
        { L"CHANNEL_32B",   AUDMODE_CHANNEL_32B,    L"Per channel 32-bit" },
        { L"SERVICE_16B",   AUDMODE_SERVICE_16B,    L"Per service 16-bit" },
        { L"SERVICE_24B",   AUDMODE_SERVICE_24B,    L"Per service 24-bit" },
        { L"SERVICE_32B",   AUDMODE_SERVICE_32B,    L"Per service 32-bit" },
        OPT_PAIR_END,
    };

    // Table defining the command-line options of DtaCmd
    static DtOptItem  CMD_OPTIONS_DEF[] =
    {
        DtOptItem(L"t", m_OptCardType, DEF_CARD_TYPE, L"Card type", 0, 5000),
        DtOptItem(L"n", m_OptCardNo, DEF_CARD_NO, L"Card number", 0, 16),
        DtOptItem(L"p", m_OptInPort, DEF_IN_PORT, L"Input port", 1, 16),
        DtOptItem(L"v", m_OptInVidStd, DEF_IN_VIDSTD, L"Video standard", 
                                                                         CMD_VIDSTD_OPTS),
        DtOptItem(L"a", m_OptAudMode, DEF_AUDIO_MODE, L"Audio mode", CMD_AUDMODE_OPTS),
    };
    static const int  NUM_CMD_OPTIONS = sizeof(CMD_OPTIONS_DEF) / 
                                                               sizeof(CMD_OPTIONS_DEF[0]);
    try
    {
        // Actual parsing of the command line options
        std::list<std::wstring>  FreeArgs;
        DtOptItem::ParseOpt(CMD_OPTIONS_DEF, argc, argv, FreeArgs);
    }
    catch (DtOptException  e)
    {
        MX_THROW_EXC_NO_ERR("Command-line error: %ls", (const wchar_t*)e);
    }
}

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- MxAvRecorderDemo::Init -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
bool  MxAvRecorderDemo::Init(DtDevice&  TheCard)
{
    DTAPI_RESULT  dr;

    //-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Prep card for matrix -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

    m_VidStd = m_LinkStd = -1;

    // The card must have been attached
    if (!TheCard.IsAttached())
        return false;
    if (!PrepCard(TheCard))
        return false;

    //+=+=+=+=+=+=+=+=+=+=+=+=+=+=+ Configure the input row +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=

    // Step 1: Attach row to the input
    DtMxPort  InPort(m_VidStd, m_LinkStd);
    dr = InPort.AddPhysicalPort(&TheCard, IN_PORT());
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed to add port %d as physical in-port", IN_PORT());
    dr = m_TheMatrix.AttachRowToInput(IN_ROW, InPort);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed to attach row %d to in-port %d", IN_ROW, IN_PORT());

    // Step 2: configure the row
    DtMxRowConfig  RowConfig;
    RowConfig.m_Enable = true;
    RowConfig.m_RowSize = 1;       // Keep a history of one frame

    // Enable the video and get scaled full picture
    RowConfig.m_VideoEnable = true;
    RowConfig.m_Video.m_PixelFormat = DT_PXFMT_UYVY422_8B; // Native packed format in 8B
    RowConfig.m_Video.m_StartLine1 = RowConfig.m_Video.m_StartLine2 = 1;
    RowConfig.m_Video.m_NumLines1 = RowConfig.m_Video.m_NumLines2 = -1;
    RowConfig.m_Video.m_Scaling = DTAPI_SCALING_1_16;   // 1/16 scaling
    RowConfig.m_Video.m_LineAlignment = 8; // Want an 8-byte alignment

    // Enable audio 
    RowConfig.m_AudioEnable = true;
    
    // We want to de-embed all available audio channels
    RowConfig.m_AudioDef.m_DeEmbed = true;
    RowConfig.m_AudioDef.m_Format = DT_AUDIO_SAMPLE_PCM;  // De-embed as PCM samples
    // m_OutputMode is irrelevant for input-only rows

    // Disable aux data
    RowConfig.m_AuxDataEnable = false;

    // Apply the row configuration
    dr = m_TheMatrix.SetRowConfig(IN_ROW, RowConfig);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed set row %d config", IN_ROW);

    // Step 3: Final configuration

    // Register our callback
    dr = m_TheMatrix.AddMatrixCbFunc(MxAvRecorderDemo::OnNewFrame, this);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed register audio callback function");

    return true;
}

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- MxAvRecorderDemo::PrepCard -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
bool  MxAvRecorderDemo::PrepCard(DtDevice&  TheCard)
{
    DTAPI_RESULT  dr;
           
    //-.-.-.-.-.-.-.-.-.-.-.-.-.- Configure port io-direction -.-.-.-.-.-.-.-.-.-.-.-.-.-.

    DtIoConfig  Cfg;
    
    Cfg.m_Port = IN_PORT();
    Cfg.m_Group = DTAPI_IOCONFIG_IODIR;
    Cfg.m_Value = DTAPI_IOCONFIG_INPUT;
    Cfg.m_SubValue = DTAPI_IOCONFIG_INPUT;
    Cfg.m_ParXtra[0] = Cfg.m_ParXtra[1] = -1;
    
    dr = TheCard.SetIoConfig(&Cfg, 1);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed to apply IO-DIR for port");

    //-.-.-.-.-.-.-.-.-.-.-.-.-.- Configure port io-standard -.-.-.-.-.-.-.-.-.-.-.-.-.-.

    // Should we auto detect? 
    m_VidStd = IN_VIDSTD(); m_LinkStd = -1;
    if (m_VidStd == DTAPI_VIDSTD_AUTO)
    {
        // Use input status object to detect video standard
        DtAvInputStatus  AvInp;
        dr = AvInp.AttachToPort(&TheCard, IN_PORT());
        do
        {
            DtDetVidStd  Info;
            dr = AvInp.DetectVidStd(Info);
            if (dr == DTAPI_OK)
            {
                m_VidStd=Info.m_VidStd; m_LinkStd=Info.m_LinkStd;
            }
            Sleep(10);

            if (_kbhit())
            {
                _getch();   // read keystroke from console
                return false;
            } 
        }
        while (dr==DTAPI_OK && m_VidStd==DTAPI_VIDSTD_UNKNOWN);
       
        printf("  Detected video standard = %s\n", ::DtapiVidStd2Str(m_VidStd));
    }
    int  IoStdValue=-1, IoStdSubValue=-1;    
    dr = ::DtapiVidStd2IoStd(m_VidStd, m_LinkStd, IoStdValue, IoStdSubValue);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed to deduce IO-STD from video and link standard");
    
    Cfg.m_Port = IN_PORT();
    Cfg.m_Group = DTAPI_IOCONFIG_IOSTD;
    Cfg.m_Value = IoStdValue;
    Cfg.m_SubValue = IoStdSubValue;
    Cfg.m_ParXtra[0] = Cfg.m_ParXtra[1] = -1;
    
    dr = TheCard.SetIoConfig(&Cfg, 1);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed to apply IO-STD for port");

    return (dr == DTAPI_OK);
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- MxAvRecorderDemo::Start -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
bool  MxAvRecorderDemo::Start()
{
    // Cannot start twice
    if (m_pRecThread != NULL)
        return false;

    //-.-.-.-.-.-.-.-.-.-.-.-.-.- Init AV streams and buffers -.-.-.-.-.-.-.-.-.-.-.-.-.-.

    // Step 1: close video streams
    for (int  i=0; i<(int)m_VidStreams.size(); i++)
        m_VidStreams[i]->Close();
    for (int  i=0; i<(int)m_AudStreams.size(); i++)
        m_AudStreams[i]->Close();

    // Step 2: reset buffers
    AvStreamBufList::iterator  it = m_VidBuffers.begin();
    for ( ; it!=m_VidBuffers.end(); it++)
        it->Clear();
    it = m_AudBuffers.begin();
    for ( ; it!=m_AudBuffers.end(); it++)
        it->Clear();

    // Step 3: clear record queue
    m_RecQueue.clear();

    //.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Init record thread -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.

    // Create state lock
    m_pRecLock = new MxCritSec;
    if (!m_pRecLock->Init())
    {
        delete m_pRecLock; m_pRecLock=NULL;
    }

    // Create record event
    m_pRecEvent = new MxEvent;
    if (!m_pRecEvent->Init())
    {
        m_pRecLock->Close();
        delete m_pRecLock; m_pRecLock=NULL;
        delete m_pRecEvent; m_pRecEvent = NULL;
        return false;
    }

    // Create record thread
    m_Exit = false;
    m_pRecThread = new MxThread;
    if (!m_pRecThread->Create(MxAvRecorderDemo::RecordLoopEntry, this))
    {
        m_pRecLock->Close();
        delete m_pRecLock; m_pRecLock=NULL;
        m_pRecEvent->Close();
        delete m_pRecEvent; m_pRecEvent = NULL;
        delete m_pRecThread; m_pRecThread = NULL;
        return false;
    }
    return MxDemoMatrixBase::Start();
}

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- MxAvRecorderDemo::Stop -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
void  MxAvRecorderDemo::Stop()
{
    MxDemoMatrixBase::Stop();

    // Clean-up record thread
    if (m_pRecThread != NULL)
    {
        m_Exit = true;
        m_pRecThread->WaitFinished();
        delete m_pRecThread; m_pRecThread = NULL;
    }
    // Clean-up record event
    if (m_pRecEvent != NULL)
    {
        m_pRecEvent->Close();
        delete m_pRecEvent; m_pRecEvent = NULL;
    }
    // Clean-up state lock
    if (m_pRecLock != NULL)
    {
        m_pRecLock->Close();
        delete m_pRecLock; m_pRecLock=NULL;
    }
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.- MxAvRecorderDemo::OnNewFrame -.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
void  MxAvRecorderDemo::OnNewFrame(DtMxData* pData, void* pOpaque)
{
    ((MxAvRecorderDemo*)pOpaque)->OnNewFrame(pData);
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.- MxAvRecorderDemo::OnNewFrame -.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
void  MxAvRecorderDemo::OnNewFrame(DtMxData* pData)
{
    // Get the frame data and check it is valid
    DtMxFrame*  pTheFrame = pData->m_Rows[IN_ROW].m_CurFrame;
    if (pTheFrame->m_Status != DT_FRMSTATUS_OK)
    {
        printf("[Frame %lld] frame status (%d) is not ok => skipping frame\n", 
                                                     pData->m_Frame, pTheFrame->m_Status);
        return;
    }

    // Get properties for video standard
    DtVidStdInfo  VidInfo;
    DTAPI_RESULT  dr = ::DtapiGetVidStdInfo(pTheFrame->m_VidStd, m_LinkStd, VidInfo);
    
    //.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Process video -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
    bool  AddedFrameToQueue = false;
    if (pTheFrame->m_VideoValid)
    {
        // Get a free buffer
        AvStreamBuf*  pAvBuf = GetFreeBuf(m_VidBuffers);
        if (pAvBuf != NULL)
        {
            MX_ASSERT(pAvBuf->m_pBuf!=NULL && pAvBuf->m_Size>0);
            pAvBuf->m_pStream = m_VidStreams[0];
            pAvBuf->m_NumValid = 0;
            pAvBuf->m_Frame = pData->m_Frame;   // Set the frame number

            // Init AV-stream video properties to zero
            pAvBuf->m_pStream->m_Width = pAvBuf->m_pStream->m_Height = 0;
                        
            unsigned char*  pDest = pAvBuf->m_pBuf;
            // Do we have an interlaced or progessive format (i.e. 1 or 2 field)?
            int  NumFields = (VidInfo.m_IsInterlaced) ? 2 : 1;
            for(int i=0; i<NumFields; i++)
            {
                const DtMxVideoBuf&  Video = pTheFrame->m_Video[i];
                // Assume a packed and planar YUV format with 8-bit symbols
                MX_ASSERT(Video.m_PixelFormat==DT_PXFMT_UYVY422_8B 
                                            || Video.m_PixelFormat==DT_PXFMT_YUYV422_8B 
                                            || Video.m_PixelFormat==DT_PXFMT_YUV422P_8B
                                            || Video.m_PixelFormat==DT_PXFMT_YUV422P2_8B
                                            || Video.m_PixelFormat==DT_PXFMT_YUV420P2_8B);

                // Update AV-stream properties with field size
                pAvBuf->m_pStream->m_Width = Video.m_Width;
                pAvBuf->m_pStream->m_Height += Video.m_Height;

                for (int j=0; j<Video.m_NumPlanes; j++)
                {
                    const DtMxVideoPlaneBuf&  Plane = Video.m_Planes[j];
                    
                    // Size of line (1-pixel = 2-symbols). NOTE: assume 8-bit symbols
                    int  LineSize = Video.m_Planes[j].m_WidthInSymbols;

                    // Copy line for line (excluding the stride padding)
                    unsigned char* pSrc = Plane.m_pBuf;
                    for (int  l=0; l<Plane.m_NumLines; l++)
                    {
                        memcpy(pDest, pSrc, LineSize);
                        MX_ASSERT((pAvBuf->m_NumValid + LineSize)<=pAvBuf->m_Size);
                        pAvBuf->m_NumValid += LineSize;
                        
                        // Move to next line
                        pDest += LineSize;
                        pSrc += (Plane.m_Stride>0) ? Plane.m_Stride : LineSize;
                    }
                }
            }

            // Add buffer to record queue
            m_pRecLock->Lock();
            m_RecQueue.push_back(pAvBuf);
            m_pRecLock->Unlock();

            AddedFrameToQueue = true;
        }
        else
        {
            printf("[Frame %lld] no free video buffer => dropping video!!\n",
                                                                          pData->m_Frame);
        }
    }

    //.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Process audio -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

    if (pTheFrame->m_AudioValid)
    {
        // Record as service or individual channels?
        if (AUDMODE()==AUDMODE_SERVICE_16B || AUDMODE()==AUDMODE_SERVICE_24B 
                                                     || AUDMODE()==AUDMODE_SERVICE_32B)
        {
            const int  SAMPLE_SIZE = (AUDMODE()==AUDMODE_SERVICE_16B) ? 16 : 
                                             ((AUDMODE()==AUDMODE_SERVICE_24B) ? 24 : 32);

            // Record as audio services (interleave audio channels)
            DtFixedVector<DtMxAudioService>&  Services = pTheFrame->m_Audio.m_Services;
            for (int  i=0; i<(int)Services.size(); i++)
            {
                // Is this service valid?
                if (!Services[i].m_Valid)
                    continue;

                AvStreamBuf*  pAvBuf = GetFreeBuf(m_AudBuffers);
                if (pAvBuf != NULL)
                {
                    MX_ASSERT(pAvBuf->m_pBuf!=NULL && pAvBuf->m_Size>0);
                    pAvBuf->m_pStream = m_AudStreams[i];
                    pAvBuf->m_NumValid = 0;
                    pAvBuf->m_Frame = pData->m_Frame;   // Set the frame number
                    pAvBuf->m_pStream->m_SampleSize = SAMPLE_SIZE;
                    pAvBuf->m_pStream->m_NumChannels = (int)Services[i].m_Channels.size();

                    // Get the audio smaples for this service as interleaved samples
                    int  NumSamples = pAvBuf->m_Size / (SAMPLE_SIZE / 8);
                    dr = pTheFrame->m_Audio.GetAudio(Services[i], pAvBuf->m_pBuf, 
                                                                NumSamples, SAMPLE_SIZE);
                    if (dr == DTAPI_OK)
                        pAvBuf->m_NumValid = NumSamples * (SAMPLE_SIZE / 8);
                    else
                    {
                        printf("[Frame %lld] failed to get audio => dropping audio!!\n",
                                                                          pData->m_Frame);
                        break;  // Skip all other services as well
                    }
                    
                    // Add buffer to record queue
                    m_pRecLock->Lock();
                    m_RecQueue.push_back(pAvBuf);
                    m_pRecLock->Unlock();

                    AddedFrameToQueue = true;
                }
                else
                {
                    printf("[Frame %lld] no free audio buffer => dropping audio!!\n",
                                                                          pData->m_Frame);
                    break;  // Skip all other services as well
                }
            }
        }
        else
        {
            // Record as individual audio channels
            DtFixedVector<DtMxAudioChannel>&  Channels = pTheFrame->m_Audio.m_Channels;
            for (int  i=0; i<(int)Channels.size(); i++)
            {
                // Do we have data for this channel?
                if (!Channels[i].m_Present || Channels[i].m_NumValidSamples==0)
                    continue;
                AvStreamBuf*  pAvBuf = GetFreeBuf(m_AudBuffers);
                if (pAvBuf != NULL)
                {
                    MX_ASSERT(pAvBuf->m_pBuf!=NULL && pAvBuf->m_Size>0);
                    pAvBuf->m_pStream = m_AudStreams[Channels[i].m_Index];
                    pAvBuf->m_NumValid = 0;
                    pAvBuf->m_Frame = pData->m_Frame;   // Set the frame number
                    pAvBuf->m_pStream->m_SampleSize= 32;  // Always 32-bit
                    pAvBuf->m_pStream->m_NumChannels = 1;   // Always one channel

                    // Copy audio samples to buffer (NOTE: each sample is 32-bit)
                    int  NumToCopy = (Channels[i].m_NumValidSamples*4);
                    memcpy(pAvBuf->m_pBuf, Channels[i].m_pBuf, NumToCopy);
                    MX_ASSERT((pAvBuf->m_NumValid + NumToCopy)<=pAvBuf->m_Size);
                    pAvBuf->m_NumValid = NumToCopy;
            
                    // Add buffer to record queue
                    m_pRecLock->Lock();
                    m_RecQueue.push_back(pAvBuf);
                    m_pRecLock->Unlock();

                    AddedFrameToQueue = true;
                }
                else
                {
                    printf("[Frame %lld] no free audio buffer => dropping audio!!\n",
                                                                          pData->m_Frame);
                    break;  // Skip all other channels as well
                }
            }
        }
    }

    // Notify record thread there is a frame to process 
    if (AddedFrameToQueue)
        m_pRecEvent->Set();
}

//.-.-.-.-.-.-.-.-.-.-.-.-.- MxAvRecorderDemo::RecordLoopEntry -.-.-.-.-.-.-.-.-.-.-.-.-.-
//
void  MxAvRecorderDemo::RecordLoopEntry(void*  pContext)
{
    ((MxAvRecorderDemo*)pContext)->RecordLoop();
}

//.-.-.-.-.-.-.-.-.-.-.-.-.-.- MxAvRecorderDemo::RecordLoop -.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
void  MxAvRecorderDemo::RecordLoop()
{
    while (!m_Exit)
    {
        // Wait for new data event
        m_pRecEvent->WaitTimeout(250);

        //.-.-.-.-.-.-.-.-.-.-.- Process all buffer in record queue -.-.-.-.-.-.-.-.-.-.-.

        // Get first buffer from queue
        m_pRecLock->Lock();
        AvStreamBuf*  pAvBuf = m_RecQueue.empty() ? NULL : m_RecQueue.front();
        m_pRecLock->Unlock();

        if (pAvBuf == NULL)
            continue;   // No buffer to process

        do
        {
            //.-.-.-.-.-.-.-.-.-.-.-.-.- Write buffer to stream -.-.-.-.-.-.-.-.-.-.-.-.-.

            // Get associated stream
            AvStream*  pAvStream = pAvBuf->m_pStream;
            MX_ASSERT(pAvStream != NULL);

            // Does the stream file exist already
            if (!pAvStream->IsOpen())
            {
                char  Filename[256];
                if (pAvStream->Type() == AvStream::VIDEO_STREAM)
                    sprintf(Filename, VIDEO_FILENAME_FMT, pAvStream->m_Width, 
                                                                     pAvStream->m_Height);
                else
                {
                    // Do we record per channel or per service
                    if (AUDMODE()==AUDMODE_CHANNEL_32B)
                    {
                        sprintf(Filename, AUDIO_FILENAME_FMT1, pAvStream->Index()+1, 
                                                                 pAvStream->m_SampleSize);
                    }
                    else
                    {
                        sprintf(Filename, AUDIO_FILENAME_FMT2, pAvStream->Index()+1,
                                       pAvStream->m_NumChannels, pAvStream->m_SampleSize);
                    }
                    
                }
                pAvStream->Open(Filename);
            }
            // Append buffer to file
            pAvStream->Write(pAvBuf->m_pBuf, pAvBuf->m_NumValid);

            //.-.-.-.-.-.-.-.-.-.-.- Update queue and buffer status -.-.-.-.-.-.-.-.-.-.-.

            // Get the lock first
            m_pRecLock->Lock();

            m_RecQueue.pop_front();     // Remove buffer from queue
            pAvBuf->m_pStream = NULL;   // Mark buffer as free by setting stream to NULL

            // Check for more buffer in queue
            pAvBuf = m_RecQueue.empty() ? NULL : m_RecQueue.front();

            m_pRecLock->Unlock();

        } while (pAvBuf!=NULL && !m_Exit);
    }
}
