//*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#* DtMxFrmBufPlayerDemo.cpp *#*#*#*#*# (C) 2015-2017 DekTec
//
// This example application demonstrates a player application based on the DtFrameBuffer 
// class from the 'old' Matrix API 1.0. For new developments it is advised to use the 
// new Matrix API 2.0 API (DtMxXxx classes). 
//
// The application generates a image (red-white-blue flag) with orange floating rectangles
// overlayed and generates 4 audio channel with a fixed tone. It also demostartes how to 
// setup your card for generatingh a genlocked output.
//
// NOTE: 
//   This application supports the DTA-2152/2154/2174
//

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Include files -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
#include <cstdio>
#include <cstring>
#include <cstdlib>
#include <vector>
#include <list>
#ifdef _WIN32
    // Windows specific includes
    #include <windows.h>
    #include <conio.h>
#else
    // Linux specific includes
    #include <unistd.h>
    #include <limits.h>
    #include <termios.h>
    #include <sys/select.h>
#endif
#include "MxFrmBufPlayerDemo.h"

// All code resides in MxExamples namespace
using namespace MxExamples;

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Application Version -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
#define APP_NAME                "DtMxFrmBufPlayerDemo"
#define APP_VERSION_MAJOR       1
#define APP_VERSION_MINOR       3
#define APP_VERSION_BUGFIX      0

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Application configuration -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
// Default, card to attach to
static const int  DEF_CARD_TYPE = 2152;
// Default, number of card to attach to
static const int  DEF_CARD_NO = 0;
// Default, output port
static const int  DEF_OUT_PORT=2;
// Genlock port and video standard. NOTE: -1, means no genlocking
static const int  DEF_GENL_PORT = -1;
static const int  DEF_GENL_VIDSTD = DTAPI_VIDSTD_525I59_94;

// Default output video standard
static const int  DEF_OUT_VIDSTD = DTAPI_VIDSTD_1080I59_94;

// Number of floating rectagles
static const int  NUM_RECTANGLES = 10;

//=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+ Main loop +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- main -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
int main(int  argc, char*  argv[])
{

    DTAPI_RESULT  dr;
    DtDevice  TheCard;
    MxFrmBufPlayerDemo  P;      // The player 

#ifndef _WIN32
    ChangeTerminalMode(1);  // Set terminal mode, so that we can use _kbhit()
#endif

    try
    {
        //-.-.-.-.-.-.-.-.-.-.-.-.-.- Display welcome message -.-.-.-.-.-.-.-.-.-.-.-.-.-.

        printf("%s V%d.%d.%d (c) 2015 DekTec Digital Video B.V.\n\n", APP_NAME,
                                APP_VERSION_MAJOR, APP_VERSION_MINOR, APP_VERSION_BUGFIX);

        //.-.-.-.-.-.-.-.-.-.-.-.- Parse the command-line options -.-.-.-.-.-.-.-.-.-.-.-.

        P.ParseCommandLine(argc, argv);        

        //.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Attach to the card -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.

        dr = TheCard.AttachToType(P.CARD_TYPE(), P.CARD_NO());
        if (dr != DTAPI_OK)
            MX_THROW_EXC(dr, "Failed to attach to DTA-%d", P.CARD_TYPE());

        // Check the port for CAP_MATRIX. DtFrameBuffer class in only support by ports
        // with this capability
        if ((TheCard.m_pHwf[(P.OUT_PORT()-1)].m_Flags&DTAPI_CAP_MATRIX) == 0)
            MX_THROW_EXC_NO_ERR("Port %d does not have the MATRIX capability", 
                                                                            P.OUT_PORT());

        printf("  Using DTA-%d, Port=%d, Standard=%s\n", P.CARD_TYPE(),
                                         P.OUT_PORT(), ::DtapiVidStd2Str(P.OUT_VIDSTD()));
        if (P.GENL_PORT() != -1)
            printf("  GenRef Port=%d, Standard=%s\n\n", P.GENL_PORT(), 
                                                      ::DtapiVidStd2Str(P.GENL_VIDSTD()));
                       
        //.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Create the player -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

        // Init the frame buffer
        if (!P.Init(TheCard))
            MX_THROW_EXC_NO_ERR("Failed to init demo");
        // Start the frame buffer
        if (!P.Start())
            MX_THROW_EXC_NO_ERR("Failed to start demo");

        //-.-.-.-.-.-.-.-.-.- Process loop, wait for user to press key -.-.-.-.-.-.-.-.-.-
        
        printf("\nPress any key to exit\n\n");
        while(!_kbhit()) { Sleep(10); }
        _getch();   // Read the keystroke from the console

        P.Stop(); // Stop the frame buffer

    }
    catch(MxDemoExc  e)
    {
        puts(e.ErrorStr());
    }
    catch(...) 
    {
        puts("Encountered an unhandled exception!!!");
    }

#ifndef _WIN32
    ChangeTerminalMode(0);  // Restore terminal mode
#endif
    return 0;
}


//=+=+=+=+=+=+=+=+=+=+=+=+=+ MxFrmBufPlayerDemo implementation +=+=+=+=+=+=+=+=+=+=+=+=+=+

//-.-.-.-.-.-.-.-.-.-.-.-.-.- MxFrmBufPlayerDemo::ThreadEntry -.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
void  MxFrmBufPlayerDemo::ThreadEntry(void*  pContext)
{
    ((MxFrmBufPlayerDemo*)pContext)->ProcessLoop();
}

//.-.-.-.-.-.-.-.-.-.-.-.- MxFrmBufPlayerDemo::MxFrmBufPlayerDemo -.-.-.-.-.-.-.-.-.-.-.-.
//
MxFrmBufPlayerDemo::MxFrmBufPlayerDemo() 
    : m_pTheThread(NULL), 
      m_Exit(false),
      m_pBackground(NULL),
      VIDEO_BUF_SIZE(4*1024*1024),
      m_pAudio(NULL),
      AUDIO_BUF_SIZE(16*1024)
{

    // Hardware requires that all buffers used for DMA are aligned on 32 byte. 
    // NOTE: only field buffers are used in DMA, hence only allocate them aligned
    for (int  i=0; i<2; i++)
        m_pField[i] = (unsigned char*)MX_UTIL::MallocAligned(32, VIDEO_BUF_SIZE);
    // Create buffer for background
    m_pBackground = new unsigned char [VIDEO_BUF_SIZE];
    // Create audio buffer
    m_pAudio = new unsigned char [AUDIO_BUF_SIZE];
}

//-.-.-.-.-.-.-.-.-.-.-.-.- MxFrmBufPlayerDemo::ParseCommandLine -.-.-.-.-.-.-.-.-.-.-.-.-
//
void  MxFrmBufPlayerDemo::ParseCommandLine(int argc, char* argv[])
{
    // Table defining video standard command-line options
    static const DtEnumOptPair CMD_VIDSTD_OPTS[] = 
    {
        // 1080: Interlaced
        { L"1080i60",    DTAPI_VIDSTD_1080I60,      L"1080i/60" },
        { L"1080i59",    DTAPI_VIDSTD_1080I59_94,   L"1080i/59.94" },
        { L"1080i50",    DTAPI_VIDSTD_1080I50,      L"1080i/50" },
        
        // 1080: progressive
        { L"1080p60",    DTAPI_VIDSTD_1080P60,      L"1080p/60" },
        { L"1080p59",    DTAPI_VIDSTD_1080P59_94,   L"1080p/59.94" },
        { L"1080p50",    DTAPI_VIDSTD_1080P50,      L"1080p/50" },
        { L"1080p30",    DTAPI_VIDSTD_1080P30,      L"1080p/30" },
        { L"1080p29",    DTAPI_VIDSTD_1080P29_97,   L"1080p/29.97" },
        { L"1080p25",    DTAPI_VIDSTD_1080P25,      L"1080p/25" },
        { L"1080p24",    DTAPI_VIDSTD_1080P24,      L"1080p/24" },
        { L"1080p23",    DTAPI_VIDSTD_1080P23_98,   L"1080p/23.98" },

        // 720: progressive
        { L"720p60",     DTAPI_VIDSTD_720P60,       L"720p/60" },
        { L"720p59",     DTAPI_VIDSTD_720P59_94,    L"720p/59.94" },
        { L"720p50",     DTAPI_VIDSTD_720P50,       L"720p/50" },
        { L"720p30",     DTAPI_VIDSTD_720P30,       L"720p/30" },
        { L"720p29",     DTAPI_VIDSTD_720P29_97,    L"720p/29.97" },
        { L"720p25",     DTAPI_VIDSTD_720P25,       L"720p/25" },
        { L"720p24",     DTAPI_VIDSTD_720P24,       L"720p/24" },
        { L"720p23",     DTAPI_VIDSTD_720P23_98,    L"720p/23.98" },

        // SD: NTSC & PAL
        { L"525i59",     DTAPI_VIDSTD_525I59_94,    L"525i/59.94" },
        { L"625i50",     DTAPI_VIDSTD_625I50,       L"625i/50" },

        OPT_PAIR_END,
    };

    // Table defining the command-line options of DtaCmd
    static DtOptItem  CMD_OPTIONS_DEF[] =
    {
        DtOptItem(L"t", m_OptCardType, DEF_CARD_TYPE, L"Card type", 0, 5000),
        DtOptItem(L"n", m_OptCardNo, DEF_CARD_NO, L"Card number", 0, 16),
        DtOptItem(L"p", m_OptOutPort, DEF_OUT_PORT, L"Output port", 1, 16),
        DtOptItem(L"v", m_OptOutVidStd, DEF_OUT_VIDSTD, L"Video standard", 
                                                                         CMD_VIDSTD_OPTS),
        DtOptItem(L"gp", m_OptGenlPort, DEF_GENL_PORT, L"Genlock reference port", -1, 16),
        DtOptItem(L"gv", m_OptGenlVidStd, DEF_GENL_VIDSTD,
                                    L"Genlock  reference video standard", CMD_VIDSTD_OPTS)
    };
    static const int  NUM_CMD_OPTIONS = sizeof(CMD_OPTIONS_DEF) / 
                                                               sizeof(CMD_OPTIONS_DEF[0]);
    try
    {
        // Actual parsing of the command line options
        std::list<std::wstring>  FreeArgs;
        DtOptItem::ParseOpt(CMD_OPTIONS_DEF, argc, argv, FreeArgs);
    }
    catch (DtOptException  e)
    {
        MX_THROW_EXC_NO_ERR("Command-line error: %ls", (const wchar_t*)e);
    }
}

//-.-.-.-.-.-.-.-.-.-.-.- MxFrmBufPlayerDemo::~MxFrmBufPlayerDemo -.-.-.-.-.-.-.-.-.-.-.-.
//
MxFrmBufPlayerDemo::~MxFrmBufPlayerDemo()
{
    Stop();

    // Free aligned field buffers
    for (int i=0; i<2; i++)
        MX_UTIL::FreeAligned(m_pField[i]);
    // Free background buffer
    delete m_pBackground;
    // Free audio buffer(s)
    delete m_pAudio;
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- MxFrmBufPlayerDemo::Init -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
bool  MxFrmBufPlayerDemo::Init(DtDevice&  TheCard)
{
    DTAPI_RESULT  dr;

    //-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Prep card for matrix -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

    // The card must have been attached
    if (!TheCard.IsAttached())
        return false;
    if (!PrepCard(TheCard))
        return false;

    //.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Attach to the port -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.

    dr = m_FrmBuf.AttachToOutput(&TheCard, OUT_PORT(), 0);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed to attach to out-port %d", OUT_PORT());

    //-.-.-.-.-.-.-.-.-.- Init our floating rectangles and background -.-.-.-.-.-.-.-.-.-.

    // Set bounding rectangle to size of image
    m_BoundingRect = MxRectangle(0,0, m_VidInfo.m_VidWidth, m_VidInfo.m_VidHeight,
                                                                            Color(0,0,0));
    
    // Seed random generator
    srand(NUM_RECTANGLES);
    MxPoint  Centre = m_BoundingRect.Centre();
    for (int  i=0; i<NUM_RECTANGLES; i++)
    {
        // Randomise size of the rectangle
        // NOTE: the x2 is because we want all our sizes to be multiples of 2
        int  Size = (((rand() % 80)*2) + 40) * 2;

        // Clip size to a maximum of 1/3th of the video height
        if (Size > (m_BoundingRect.Height() / 3))
            Size = (m_BoundingRect.Height() / 6) * 2;

        // Create the a rectangle with the specified size
        MxRectangle  Rect = MxRectangle(0, 0, Size, Size, COL_ORANGE);

        // Start at the centre
        MxPoint  P = Centre;
        P.Offset(-Size/2, -Size/2);
        Rect.MoveTo(P);

        // Randomise movement vector
        int  SpeedX = (rand()%6 + 1) * 2;
        int  SpeedY = (rand()%8 + 1) * 2;
        // Randomise direction
        switch (rand()%4)
        {
        // Move left and upwards
        case 0:  Rect.m_Vector = MxPoint(-SpeedX, -SpeedY); break;
        // Move left and downwards
        case 1:  Rect.m_Vector = MxPoint(-SpeedX, SpeedY); break;
        // Move right and upwards
        case 2:  Rect.m_Vector = MxPoint(SpeedX, -SpeedY); break;
        // Move right and downwards
        case 3:  Rect.m_Vector = MxPoint(SpeedX, SpeedY); break;
        }

        // Add MxRectangle to list
        m_Rectangles.push_back(Rect);
    }

    // Init background buffer
    unsigned short*  pPixel = (unsigned short*)m_pBackground;
    for (int  i=0; i<m_VidInfo.m_VidHeight; i++)
    {
        //-.-.-.-.-.-.-.-.-.- Generate red, white and blue back ground -.-.-.-.-.-.-.-.-.-

        Color  Col(0,0,0);
        if (i<(m_VidInfo.m_VidHeight/3))            Col = COL_RED;
        else if (i<((m_VidInfo.m_VidHeight/3)*2))   Col = COL_WHITE;
        else                                        Col = COL_BLUE;

        // Write two pixels at a time
        for (int  j=0; j<m_VidInfo.m_VidWidth; j+=2)
        {
            *pPixel++ = Col.Pixel1(); *pPixel++ = Col.Pixel2();
        }
    }
    return true;
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.- MxFrmBufPlayerDemo::PrepCard -.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
bool  MxFrmBufPlayerDemo::PrepCard(DtDevice&  TheCard)
{
    DTAPI_RESULT  dr;

    // Get properties (e.g. dimensions, fractional, interlaced) for output video standard
    dr = ::DtapiGetVidStdInfo(OUT_VIDSTD(), m_VidInfo);

    //+=+=+=+=+=+=+=+=+=+=+=+ Compute and apply IO-configurations +=+=+=+=+=+=+=+=+=+=+=+=
    
    DtIoConfig  Cfg;
    std::vector<DtIoConfig>  IoConfigs;

    //-.-.-.-.-.-.-.-.-.-.-.- Step 1: configure genlock reference -.-.-.-.-.-.-.-.-.-.-.-.
    // NOTE: only one port can be configured as the genlock referenec, therefore we must
    // disable the GENREF IO-config on all port but the 'one'

    // Step 1.1: Compute GENREF io-configurations
    DtHwFuncDesc*  pHwf = TheCard.m_pHwf;
    for (int  i=0; i<TheCard.m_pHwf->m_DvcDesc.m_NumPorts; i++, pHwf++)
    {
        // Check if the port has GENNREF capability
        if ((pHwf->m_Flags & DTAPI_CAP_GENREF) == 0)
            continue;   // No GENREF cap

        // Is this the internal GENREF port (has DTAPI_CAP_VIRTUAL)?
        bool  IsIntGenref = (pHwf->m_Flags & DTAPI_CAP_VIRTUAL)!=0;
        
        Cfg.m_Port = pHwf->m_Port;
        Cfg.m_Group = DTAPI_IOCONFIG_GENREF;

        // When no GENLOCK-ing is required use the internal GENREF port
        if (GENL_PORT() == -1)
            Cfg.m_Value = (IsIntGenref) ? DTAPI_IOCONFIG_TRUE : DTAPI_IOCONFIG_FALSE;
        else if (pHwf->m_Port == GENL_PORT())
            Cfg.m_Value = DTAPI_IOCONFIG_TRUE;
        else
            Cfg.m_Value = DTAPI_IOCONFIG_FALSE;
        
        Cfg.m_SubValue = -1;
        Cfg.m_ParXtra[0] = Cfg.m_ParXtra[1] = -1;
        // Add to list
        IoConfigs.push_back(Cfg);

        // If 'THE' GENREF port has input capability, make sure it is configured as input
        if (pHwf->m_Port==GENL_PORT() && (pHwf->m_Flags & DTAPI_CAP_INPUT)!=0)
        {
            Cfg.m_Group = DTAPI_IOCONFIG_IODIR;
            Cfg.m_Value = DTAPI_IOCONFIG_INPUT;
            Cfg.m_SubValue = DTAPI_IOCONFIG_INPUT;
            Cfg.m_ParXtra[0] = Cfg.m_ParXtra[1] = -1;
            // Add to list
            IoConfigs.push_back(Cfg);
        }

        // Set IO-standard of GENREF signal
        if (pHwf->m_Port==GENL_PORT() || (GENL_PORT()==-1 && IsIntGenref))
        {
            // IO-standard: derive from desired output video standard
            Cfg.m_Group = DTAPI_IOCONFIG_IOSTD;
            ::DtapiVidStd2IoStd(GENL_VIDSTD(), Cfg.m_Value, Cfg.m_SubValue);
            Cfg.m_ParXtra[0] = Cfg.m_ParXtra[1] = -1;
            // Add to list
            IoConfigs.push_back(Cfg);
        }
    }

    // Step 1.2: Apply genref io-configurations
    dr = TheCard.SetIoConfig(IoConfigs.data(), IoConfigs.size());
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed to apply output port io-configurations");

    // Step 1.3: Wait for lock on the reference input
    if (GENL_PORT() != -1)
    {
        const int  WAIT_TIME_MS=250, MAX_WAIT_TIME_MS=10000, MIN_WAIT_TIME_MS=500;
        const int  MIN_LOOP_CNT=(MIN_WAIT_TIME_MS/WAIT_TIME_MS);
        const int  MAX_LOOP_CNT=(MAX_WAIT_TIME_MS/WAIT_TIME_MS);
        int  LoopCnt=0, State=DTAPI_GENL_NO_REF;
        do
        {
            const char  ActivityIndicator[4] = {'|', '/', '-', '\\'};
            const char*  StateStr;
            if (State == DTAPI_GENL_LOCKED)         StateStr = "LOCKED";
            else if (State == DTAPI_GENL_NO_REF)    StateStr = "NO REF";
            else if (State == DTAPI_GENL_LOCKING)   StateStr = "LOCKING";
            else                                    StateStr = "NO LOCK";

            printf("%c Waiting for GenRef lock (current state=%s)  \r", 
                                                  ActivityIndicator[LoopCnt%4], StateStr);
            
            Sleep(WAIT_TIME_MS);
            
            // Get lock state
            dr = TheCard.GetGenlockState(State);
            if (dr != DTAPI_OK)
                MX_THROW_EXC(dr, "GetGenlockState failed");

            if (_kbhit())
            {
                _getch();   // read keystroke from console
                return false;
            }
            LoopCnt++;
        }
        while((State!=DTAPI_GENL_LOCKED || LoopCnt<MIN_LOOP_CNT) && LoopCnt<MAX_LOOP_CNT);

        if (State != DTAPI_GENL_LOCKED)
            printf("\nWARNING: Genlock reference port (%d) has not locked\n", 
                                                                             GENL_PORT());
    }
    
    //.-.-.-.-.-.-.-.-.-.-.-.- Step 2: IO-configs for output port -.-.-.-.-.-.-.-.-.-.-.-.

    IoConfigs.clear();  // Clear GENREF io-config from list

    // Step 2.1: Compute output io-configs
    Cfg.m_Port = OUT_PORT();

    // IO-direction: configure as output
    Cfg.m_Group = DTAPI_IOCONFIG_IODIR;
    Cfg.m_Value = DTAPI_IOCONFIG_OUTPUT;
    Cfg.m_SubValue = DTAPI_IOCONFIG_OUTPUT;
    Cfg.m_ParXtra[0] = Cfg.m_ParXtra[1] = -1;
    // Add to list
    IoConfigs.push_back(Cfg);

    // IO-standard: derive from desired output video standard
    Cfg.m_Group = DTAPI_IOCONFIG_IOSTD;
    ::DtapiVidStd2IoStd(OUT_VIDSTD(), Cfg.m_Value, Cfg.m_SubValue);
    Cfg.m_ParXtra[0] = Cfg.m_ParXtra[1] = -1;
    // Add to list
    IoConfigs.push_back(Cfg);

    // GENLOCKED: set to TRUE if a GENREF port is used
    Cfg.m_Group = DTAPI_IOCONFIG_GENLOCKED;
    Cfg.m_Value = (GENL_PORT()!=-1) ? DTAPI_IOCONFIG_TRUE : DTAPI_IOCONFIG_FALSE;
    Cfg.m_SubValue = -1;
    Cfg.m_ParXtra[0] = Cfg.m_ParXtra[1] = -1;
    // Add to list
    IoConfigs.push_back(Cfg);

    // Step 2.2: set fractional mode io-config (only needed for ports with FRACMODE cap)
    // NOTE: FRACMODE must be set the same for all ports on the boards
    pHwf = TheCard.m_pHwf;
    for (int  i=0; i<TheCard.m_pHwf->m_DvcDesc.m_NumPorts; i++, pHwf++)
    {
        // Check if the port has FRACMODE capability
        if ((pHwf->m_Flags & DTAPI_CAP_FRACMODE) == 0)
            continue;   // No FRACMODE cap

        Cfg.m_Port = pHwf->m_Port;
        Cfg.m_Group = DTAPI_IOCONFIG_FRACMODE;
        Cfg.m_Value = (m_VidInfo.m_IsFractional) ? DTAPI_IOCONFIG_TRUE : 
                                                                     DTAPI_IOCONFIG_FALSE;
        Cfg.m_SubValue = -1;
        Cfg.m_ParXtra[0] = Cfg.m_ParXtra[1] = -1;
        // Add to list
        IoConfigs.push_back(Cfg);
    }
       
    // Step 2.3: Apply io-configs
    dr = TheCard.SetIoConfig(IoConfigs.data(), IoConfigs.size());
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed to apply output port io-configurations");

    return true;
}

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- MxFrmBufPlayerDemo::Start -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
bool  MxFrmBufPlayerDemo::Start()
{
    // Cannot start twice
    if (m_pTheThread != NULL)
        return  false;

    m_Exit = false;
    m_pTheThread = new MxThread;
    return m_pTheThread->Create(MxFrmBufPlayerDemo::ThreadEntry, this);
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- MxFrmBufPlayerDemo::Stop -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
void  MxFrmBufPlayerDemo::Stop()
{
    if (m_pTheThread == NULL)
        return;

    // Kill out thread
    m_Exit = true;
    m_pTheThread->WaitFinished();

    // Clean-up thread resouces
    m_pTheThread->Close();
    delete m_pTheThread; m_pTheThread = NULL;
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.- MxFrmBufPlayerDemo::ProcessLoop -.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
void  MxFrmBufPlayerDemo::ProcessLoop()
{
    DTAPI_RESULT  dr;
    try
    {
        // Start transmission
        dr = m_FrmBuf.Start(true);
        if (dr != DTAPI_OK)
            MX_THROW_EXC(dr, "Failed to start frame-buffer");
    
        __int64  Frame=-1;
        while (!m_Exit)
        {
            // Wait for 'space' to write new frames
            __int64  FirstSafeFrame=-1, LastSafeFrame=-1;
            dr = m_FrmBuf.WaitFrame(FirstSafeFrame, LastSafeFrame);
            if (dr != DTAPI_OK)
            {
                Sleep(1);
                continue;
            }
            // Initial initilisation
            if (Frame == -1)
                Frame = FirstSafeFrame;
            
            // Check for 'underflow'
            if (Frame < FirstSafeFrame)
            {
                printf("WARNING: underflow detected [Frame=%lld < FirstSafe=%lld]\n", 
                                                                   Frame, FirstSafeFrame);
                Frame = FirstSafeFrame; // continue from first safe frame
            }

            for ( ; Frame<LastSafeFrame; Frame++)
            {
                // Write video to frame-buffer
                WriteVideo(Frame);
                // Write audio to frame buffer
                WriteAudio(Frame);
                // Commit HANC/VANC to 
                dr = m_FrmBuf.AncCommit(Frame);
                if (dr != DTAPI_OK)
                    MX_THROW_EXC(dr, "[%lld] AncCommit failed", Frame);
            }
        }
    }
    catch(MxDemoExc  e)
    {
        puts(e.ErrorStr());
    }
    catch(...) 
    {
        puts("Encountered an unhandled exception!!!");
    }

    // Stop transmission
    m_FrmBuf.Start(false);
}

//.-.-.-.-.-.-.-.-.-.-.-.-.-.- MxFrmBufPlayerDemo::WriteVideo -.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
bool  MxFrmBufPlayerDemo::WriteVideo(__int64  Frame)
{
    DTAPI_RESULT  dr;

    //.-.-.-.-.-.-.-.- Step 1: generate a single progressive video frame -.-.-.-.-.-.-.-.-
    // NOTE: the progressive frame is generated in field 1 buffer

    // Simple copy background to field 1
    memcpy(m_pField[0], m_pBackground, VIDEO_BUF_SIZE);

    //.-.-.-.-.-.-.-.-.-.-.-.-.-.- Overlay the rectangles -.-.-.-.-.-.-.-.-.-.-.-.-.-.
    for (int  i=0; i<(int)m_Rectangles.size(); i++)
    {
        MxRectangle&  R = m_Rectangles[i];

        // Move rectangle to new position
        R.Offset(R.m_Vector);

        // Bounce of the left and right edge
        if (R.Left() < m_BoundingRect.Left())
        {
            R.Offset(-R.Left()*2, 0);
            // Invert horizontal direction of movement vector
            R.m_Vector.m_X = -R.m_Vector.m_X;
        }
        else if (R.Right() > m_BoundingRect.Right())
        {
            R.Offset((m_BoundingRect.Right()-R.Right())*2, 0);
            // Invert horizontal direction of movement vector
            R.m_Vector.m_X = -R.m_Vector.m_X;
        }

        // Bounce of the top and bottom edge
        if (R.Top() < m_BoundingRect.Top())
        {
            R.Offset(0, -R.Top()*2);
            // Invert vertical direction of movement vector
            R.m_Vector.m_Y = -R.m_Vector.m_Y;
        }
        else if (R.Bottom() > m_BoundingRect.Bottom())
        {
            R.Offset(0, (m_BoundingRect.Bottom()-R.Bottom())*2);
            // Invert vertical direction of movement vector
            R.m_Vector.m_Y = -R.m_Vector.m_Y;
        }

        // Draw rectangle on top of the background
        for (int  j=0; j<R.Height(); j++)
        {
            // Computer start position
            unsigned short*  pPixel = ((unsigned short*)m_pField[0]) + 
                                                   (m_VidInfo.m_VidWidth * (R.Top() + j));
            pPixel += R.Left();
            for (int  n=0; n<R.Width(); n+=2)
            {
                *pPixel++ = R.m_Color.Pixel1(); *pPixel++ = R.m_Color.Pixel2();
            }
        }
    }

    //-.-.-.-.-.-.-.-.-.-.-.-.- Step 2: Write video to hardware -.-.-.-.-.-.-.-.-.-.-.-.-.

    // Interlaced or progressive output?
    if (!m_VidInfo.m_IsInterlaced)
    {
        //-.-.-.-.-.- Step 2.1: write our progressive field 1 to frame-buffer -.-.-.-.-.-.

        int  NumLines=-1, BufSize=VIDEO_BUF_SIZE;
        dr = m_FrmBuf.WriteVideo(Frame, m_pField[0], BufSize, DTAPI_SDI_FIELD1, 
                                                               DTAPI_SDI_8B, 1, NumLines);
        if (dr != DTAPI_OK)
            MX_THROW_EXC(dr, "[F1] WriteVideo failed");
    }
    else
    {
        //-.-.-.-.- Step 2.1: 'convert' progressive field 1 to interlaced fields -.-.-.-.-
               
        const int  LineNumS = m_VidInfo.m_VidWidth*2;
        const unsigned char *pEven=NULL, *pOdd=NULL;   
        unsigned char *pDest1=m_pField[0], *pDest2=m_pField[1];
        for (int  i=0; i<m_VidInfo.m_VidHeight; i+=2)
        {
            // Move even line to correct position within field 1
            pEven = m_pField[0] + (i*LineNumS);
            if (pDest1 != pEven)
                memmove(pDest1, pEven, LineNumS);
            pDest1 += LineNumS;

            // Copy odd line to correct position in field 2
            pOdd = m_pField[0] + ((i+1)*LineNumS);
            memcpy(pDest2, pOdd, LineNumS);
            pDest2 += LineNumS;
        }

        //.-.-.-.-.-.-.-.-.- Step 2.1: write interlaced fields to frame -.-.-.-.-.-.-.-.-.

        // Field 1
        int  NumLines=-1, BufSize=VIDEO_BUF_SIZE;
        dr = m_FrmBuf.WriteVideo(Frame, m_pField[0], BufSize, DTAPI_SDI_FIELD1, 
                                                               DTAPI_SDI_8B, 1, NumLines);
        if (dr != DTAPI_OK)
            MX_THROW_EXC(dr, "[F1] WriteVideo failed");
        // Field 2
        NumLines=-1, BufSize=VIDEO_BUF_SIZE;
        dr = m_FrmBuf.WriteVideo(Frame, m_pField[1], BufSize, DTAPI_SDI_FIELD2, 
                                                               DTAPI_SDI_8B, 1, NumLines);
        if (dr != DTAPI_OK)
            MX_THROW_EXC(dr, "[F2] WriteVideo failed");
    }
    return true;
}

//.-.-.-.-.-.-.-.-.-.-.-.-.-.- MxFrmBufPlayerDemo::WriteAudio -.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
bool  MxFrmBufPlayerDemo::WriteAudio(__int64  Frame)
{
    DTAPI_RESULT  dr;

    //-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Add PCM samples -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.

    static MxToneGenerator  TONE_GEN[4] = 
    {
        MxToneGenerator(1000, 0.4 / ((rand()%4)+1)), // 1.0KHz tone; with random volume
        MxToneGenerator(1500, 0.4 / ((rand()%4)+1)), // 1.5KHz tone; with random volume
        MxToneGenerator(2000, 0.4 / ((rand()%4)+1)), // 2.0KHz tone; with random volume
        MxToneGenerator(2500, 0.4 / ((rand()%4)+1)), // 2.5KHz tone; with random volume
    };

    // Generate audio samples and write them to frame-buffer
    int  Channel = DTAPI_SDI_AUDIO_CHAN1;
    for (int  i=0; i<4; i++, Channel<<=1)
    {
        int  BufSize=0, NumSamples=0;
        // Call AncAddAudio with dummy buffer to get number of expected sample bytes
        dr = m_FrmBuf.AncAddAudio(Frame, NULL, BufSize, DTAPI_SDI_AUDIO_PCM32, Channel, 
                                                                  DTAPI_SDI_AUDIO_GROUP1);
        if (dr != DTAPI_E_BUF_TOO_SMALL)
            MX_THROW_EXC(dr, "[C=%d] AncAddAudio failed", i+1);

        // Generate audio samples for current channel
        NumSamples = BufSize / 4;
        TONE_GEN[i].Generate((unsigned int*)m_pAudio, NumSamples);

        // Now: actually write audio sample to frame-buffer
        dr = m_FrmBuf.AncAddAudio(Frame, m_pAudio, BufSize, DTAPI_SDI_AUDIO_PCM32, 
                                                         Channel, DTAPI_SDI_AUDIO_GROUP1);
        if (dr != DTAPI_OK)
            MX_THROW_EXC(dr, "[C=%d] AncAddAudio failed", i+1);
    }
    return true;
}