//*#*#*#*#*#*#*#*#*#*#*#*#*#*#* DtMxHdmi192KHzAudioDemo.cpp *#*#*#*#*#*#*# (C) 2018 DekTec
//
// This example application demonstrates a matrix configuration for a 4k (UHD) player. The
// application uses two callbacks: one to generate the 4k image (red-white-blue flag) with
// orange floating rectangles overlayed and one to generate the audio (2x stereo, with
// a fixed tone and 2x192KHz audio).
//
// NOTE: 
//  - This application uses the HDMI output port of the DTA-2195 to generate
//    HDMI compliant output with 48KHz and 192KHz audio.

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Include files -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
#include <cstdio>
#include <cstring>
#include <cstdlib>
#include <vector>
#ifdef _WIN32
    // Windows specific includes
    #include <windows.h>
    #include <conio.h>
#else
    // Linux specific includes
    #include <unistd.h>
    #include <limits.h>
    #include <termios.h>
    #include <sys/select.h>
#define _stricmp strcasecmp
#endif
#include "MxHdmi192KHzAudioDemo.h"

// All code resides in MxExamples namespace
using namespace MxExamples;
// All code resides in MxExamples namespace

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Application Version -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
#define APP_NAME                "DtMxHdmi192KHzAudioDemo"
#define APP_VERSION_MAJOR       1
#define APP_VERSION_MINOR       0
#define APP_VERSION_BUGFIX      0

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Application configuration -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
// Card to attach to
static const int  CARD_TYPE = 2195;
// Number of card to attach to
static const int  CARD_NO = 0;
// Output ports
static const int  SDI_OUTPUT_PORT = 2;
static const int  HDMI_OUTPUT_PORT = 4;
// Row indices
static const int  OUT_ROW=0;
// Number of floating rectagles
static const int  NUM_RECTANGLES = 20;

// Video standard and link standard. Only use 4k standards/formats
int  VIDSTD = DTAPI_VIDSTD_2160P60;
int  LINKSTD = DTAPI_VIDLNK_4K_SMPTE2082;

// 2x 48khz stereo channel
static const int  NUM_48KHZ_AUDIO_CHANNELS = 4;
static const int  NUM_192KHZ_AUDIO_CHANNELS = 2;

// Enable this define to get profiling information
#define  MX_PRINT_PROFILING_INFO

static const DtMxAudioSampleType  AUDIO_MODE = DT_AUDIO_SAMPLE_AES3;
static const DtMxPixelFormat  PIXFMT = DT_PXFMT_YUV422P_8B;// Use a planar format (8-bit);

//=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+ Main loop +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- main -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
int main(int  argc, char*  argv[])
{
    DTAPI_RESULT  dr;
    DtDevice  TheCard;
    MxHdmi192KHzAudioDemo  Hdmi192KHzAudioDemo;

#ifndef _WIN32
    ChangeTerminalMode(1);  // Set terminal mode, so that we can use _kbhit()
#endif

    try
    {
        //-.-.-.-.-.-.-.-.-.-.-.-.-.- Display welcome message -.-.-.-.-.-.-.-.-.-.-.-.-.-.

        printf("%s V%d.%d.%d (c) 2018 DekTec Digital Video B.V.\n\n", APP_NAME,
                                APP_VERSION_MAJOR, APP_VERSION_MINOR, APP_VERSION_BUGFIX);

        //.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Attach to the card -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.

        dr = TheCard.AttachToType(CARD_TYPE, CARD_NO);
        if (dr != DTAPI_OK)
            MX_THROW_EXC(dr, "Failed to attach to DTA-%d", CARD_TYPE);

        // Check the ports for CAP_MATRIX2. The DtMxXxx API is only supported by ports 
        // with DTAPI_CAP_MATRIX2 capability

        if ((TheCard.m_pHwf[SDI_OUTPUT_PORT-1].m_Flags&DTAPI_CAP_MATRIX2) == 0)
        {
            MX_THROW_EXC_NO_ERR("Port %d does not have the MATRIX2 capability",
                                                TheCard.m_pHwf[SDI_OUTPUT_PORT-1].m_Port);
        }
                       
        //.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Create the matrix -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

        // Init the matrix
        if (!Hdmi192KHzAudioDemo.Init(TheCard))
            MX_THROW_EXC_NO_ERR("Failed to init demo");
        // Start the matrix
        if (!Hdmi192KHzAudioDemo.Start())
            MX_THROW_EXC_NO_ERR("Failed to start demo");

        //-.-.-.-.-.-.-.-.-.- Process loop, wait for user to press key -.-.-.-.-.-.-.-.-.-
        
        printf("\nPress t to toggle audio, any other key to stop\n\n");
        bool stop = false;
        while(!stop) 
        { 
#ifdef _WIN32
            int key = _getch();   // Read the keystroke from the console
#else
            int key = getchar();   // Read the keystroke from the console
#endif
            if (key == 't')
                Hdmi192KHzAudioDemo.ToggleAudioChannel();
            else 
                stop = true;
        }
        
        Hdmi192KHzAudioDemo.Stop(); // Stop the matrix

        // Print profling information
#ifdef MX_PRINT_PROFILING_INFO
        Hdmi192KHzAudioDemo.PrintProfilingInfo();
        printf("\nPress any key to exit\n");
        // Wait for a keystroke before exiting. To allow user time to take in the 
        // profiling information
        _getch();
#endif
    }
    catch(MxDemoExc  e)
    {
        puts(e.ErrorStr());
    }
    catch(...) 
    {
        puts("Encountered an unhandled exception!!!");
    }

#ifndef _WIN32
    ChangeTerminalMode(0);  // Restore terminal mode
#endif
    return 0;
}

namespace MxExamples
{

//+=+=+=+=+=+=+=+=+=+=+=+=+ MxHdmi192KHzAudioDemo implementation +=+=+=+=+=+=+=+=+=+=+=+=+

//-.-.-.-.-.-.-.-.-.-.- MxHdmi192KHzAudioDemo::MxHdmi192KHzAudioDemo -.-.-.-.-.-.-.-.-.-.-
//
MxHdmi192KHzAudioDemo::MxHdmi192KHzAudioDemo()
{
    m_pAudioBuffer[0] = NULL;
    m_AudioBufferLength[0] = 0;
    m_pAudioBuffer[1] = NULL;
    m_AudioBufferLength[1] = 0;
}

//.-.-.-.-.-.-.-.-.-.- MxHdmi192KHzAudioDemo::~MxHdmi192KHzAudioDemo -.-.-.-.-.-.-.-.-.-.-
//
MxHdmi192KHzAudioDemo::~MxHdmi192KHzAudioDemo()
{
    Stop(); // Call stop, just in case we did not stop it ourselves
}

void  MxHdmi192KHzAudioDemo::ToggleAudioChannel()
{
    if (!m_192kHz)
        m_AvOutput.SetHdmiAudioChannel(5, 9);
    else
        m_AvOutput.SetHdmiAudioChannel(1, 2);
    m_192kHz = !m_192kHz;
    if (m_192kHz)
        printf("  Selected 192kHz audio channel 5 and 9\n");
    else
        printf("  Selected  48kHz audio channel 1 and 2\n");
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.- MxHdmi192KHzAudioDemo::Init -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
bool  MxHdmi192KHzAudioDemo::Init(DtDevice&  TheCard)
{
    DTAPI_RESULT  dr;

    //-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Prep card for matrix -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

    // The card must have been attached
    if (!TheCard.IsAttached())
        return false;
    if (!PrepCard(TheCard))
        return false;

    if (TheCard.TypeNumber() == 2195)
    {
        dr = m_AvOutput.AttachToPort(&TheCard, HDMI_OUTPUT_PORT);
        if (dr != DTAPI_OK)
        {
            printf("  Error attaching to HDMI port\n");
            return false;
        }
    }

    //+=+=+=+=+=+=+=+=+=+=+=+=+=+=+ Configure the output row +=+=+=+=+=+=+=+=+=+=+=+=+=+=+

    //.-.-.-.-.-.-.-.-.-.-.-.-.-.- Step 1: Create the output port -.-.-.-.-.-.-.-.-.-.-.-.
    DtMxPort  OutputPort(VIDSTD, LINKSTD);
    dr = OutputPort.AddPhysicalPort(&TheCard, SDI_OUTPUT_PORT);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed to add port %d", SDI_OUTPUT_PORT);

    //-.-.-.-.-.-.-.-.-.-.-.-.- Step 2: attach the output port to a row -.-.-.-.-.-.-.-.-.
    // Set our output port as the row output
    dr = m_TheMatrix.AttachRowToOutput(OUT_ROW, OutputPort);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed to attach row %d to the 4k output port", OUT_ROW);

    printf("  Video standard = %s\n", ::DtapiVidStd2Str(VIDSTD));
    printf("  Link standard = %s\n", ::DtapiLinkStd2Str(LINKSTD));

    //.-.-.-.-.-.-.-.-.-.-.-.-.- Step 3: set row configuration -.-.-.-.-.-.-.-.-.-.-.-.-.-

    m_RowConfig.m_Enable = true;
    m_RowConfig.m_RowSize = 1;       // Keep a history of one frame

    // Enable the video
    m_RowConfig.m_VideoEnable = true;
    m_RowConfig.m_Video.m_PixelFormat = PIXFMT;
    m_RowConfig.m_Video.m_StartLine1 = m_RowConfig.m_Video.m_StartLine2 = 1;
    m_RowConfig.m_Video.m_NumLines1 = m_RowConfig.m_Video.m_NumLines2 = -1;
    m_RowConfig.m_Video.m_Scaling = DTAPI_SCALING_OFF;   // No scaling for this row
    m_RowConfig.m_Video.m_LineAlignment = 8; // Want an 8-byte alignment

    // Enable audio
    m_RowConfig.m_AudioEnable = true;
    // We will add twelf channels 2x48kHz stereo = 2x (2 channels of 48kHz)
    //                       and  2x192khz       = 2x (4 channels of 48khz)
    
    for (int  i=0; i<NUM_48KHZ_AUDIO_CHANNELS+NUM_192KHZ_AUDIO_CHANNELS*4; i++)
    {
        DtMxAudioConfig  AudioConfig;
        AudioConfig.m_Index = i;
        AudioConfig.m_DeEmbed = false;  // Nothing to de-embed for an output only row
        AudioConfig.m_OutputMode = DT_OUTPUT_MODE_ADD;  // We want to add audio samples
        AudioConfig.m_Format = AUDIO_MODE;     // We will add AES3 or PCM samples

        // Add to config list
        m_RowConfig.m_Audio.push_back(AudioConfig);
    }

    // Set start first audio output to channel 5 and second audio output channel to 9
    // to output the 2x192kHz audio channels
    m_192kHz = false;
    if (m_AvOutput.IsAttached())
    {
        ToggleAudioChannel();
    }
   
    // Apply the row configuration
    dr = m_TheMatrix.SetRowConfig(OUT_ROW, m_RowConfig);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed set row %d config", OUT_ROW);
    
    //-.-.-.-.-.-.-.-.-.-.-.-.-.- Step 4: Final configuration -.-.-.-.-.-.-.-.-.-.-.-.-.-.

    // Determine the default recommended end-to-end delay for this configuration
    int  DefEnd2EndDelay = 0;
    double CbFrames=0.0;    // Time available for the call-back method
    dr = m_TheMatrix.GetDefEndToEndDelay(DefEnd2EndDelay, CbFrames);
    printf("  Default end-to-end delay = %d Frames\n", DefEnd2EndDelay);
    printf("  Time-for-Callback = %.1f Frame Period\n", CbFrames);
    // Apply the default end-to-end delay
    dr = m_TheMatrix.SetEndToEndDelay(DefEnd2EndDelay);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed set the end-to-end delay (delay=%d)", DefEnd2EndDelay);

    // Register our callbacks
    m_ContextAudio.m_pPlayer = this; 
    m_ContextAudio.m_Type = CallbackContext::GEN_AUDIO;
    dr = m_TheMatrix.AddMatrixCbFunc(MxHdmi192KHzAudioDemo::OnNewFrame, &m_ContextAudio);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed register audio callback function");
    
    m_ContextVideo.m_pPlayer = this; 
    m_ContextVideo.m_Type = CallbackContext::GEN_VIDEO;
    dr = m_TheMatrix.AddMatrixCbFunc(MxHdmi192KHzAudioDemo::OnNewFrame, &m_ContextVideo);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed register video callback function");

    //-.-.-.-.-.-.-.-.-.-.-.-.-.- Init our floating rectangles -.-.-.-.-.-.-.-.-.-.-.-.-.-

    DtVidStdInfo  VidInfo;
    dr = ::DtapiGetVidStdInfo(VIDSTD, LINKSTD, VidInfo);

    // Set bounding rectangle to size of image
    m_BoundingRect = MxRectangle(0,0, VidInfo.m_VidWidth, VidInfo.m_VidHeight,
                                                                            Color(0,0,0));
    
    // Seed random generator
    srand(NUM_RECTANGLES);
    MxPoint  Centre = m_BoundingRect.Centre();
    for (int  i=0; i<NUM_RECTANGLES; i++)
    {
        // Randomise size of the rectangle
        // NOTE: the x2 is because we want all our sizes to be multiples of 2
        int  Size = (((rand() % 100)*2) + 50) * 2;

        // Clip size to a maximum of 1/3th of the video height
        if (Size > (m_BoundingRect.Height() / 3))
            Size = (m_BoundingRect.Height() / 6) * 2;

        // Create the a rectangle with the specified size
        MxRectangle  Rect = MxRectangle(0, 0, Size, Size, COL_ORANGE);

        // Start at the centre
        MxPoint  P = Centre;
        P.Offset(-Size/2, -Size/2);
        Rect.MoveTo(P);

        // Randomise movement vector
        int  SpeedX = (rand()%6 + 1) * 2;
        int  SpeedY = (rand()%7 + 1) * 2;
        // Randomise direction
        switch (rand()%4)
        {
        // Move left and upwards
        case 0:  Rect.m_Vector = MxPoint(-SpeedX, -SpeedY); break;
        // Move left and downwards
        case 1:  Rect.m_Vector = MxPoint(-SpeedX, SpeedY); break;
        // Move right and upwards
        case 2:  Rect.m_Vector = MxPoint(SpeedX, -SpeedY); break;
        // Move right and downwards
        case 3:  Rect.m_Vector = MxPoint(SpeedX, SpeedY); break;
        }

        // Add MxRectangle to list
        m_Rectangles.push_back(Rect);
    }
    return true;
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.- MxHdmi192KHzAudioDemo::PrepCard -.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
bool  MxHdmi192KHzAudioDemo::PrepCard(DtDevice&  TheCard)
{
    DTAPI_RESULT  dr;
           
    //-.-.-.-.-.-.-.-.-.-.-.-.-.- Configure port io-direction -.-.-.-.-.-.-.-.-.-.-.-.-.-.

    DtIoConfig  Cfg;
    std::vector<DtIoConfig>  IoConfigs;

    Cfg.m_Port = SDI_OUTPUT_PORT;
    Cfg.m_Group = DTAPI_IOCONFIG_IODIR;
    Cfg.m_Value = DTAPI_IOCONFIG_OUTPUT;
    Cfg.m_SubValue = DTAPI_IOCONFIG_OUTPUT;
    Cfg.m_ParXtra[0] = Cfg.m_ParXtra[1] = -1;
    // Add to list
    IoConfigs.push_back(Cfg);

    dr = TheCard.SetIoConfig(IoConfigs.data(), (int)IoConfigs.size());
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed to apply IO-DIR for ports");

    //-.-.-.-.-.-.-.-.-.-.-.-.-.- Configure ports io-standard -.-.-.-.-.-.-.-.-.-.-.-.-.-.
    // Configure IO-STD for all 4 ports 

    int  IoStdValue=-1, IoStdSubValue=-1;
    dr = ::DtapiVidStd2IoStd(VIDSTD, LINKSTD, IoStdValue, IoStdSubValue);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed to deduce IO-STD from video and link standard");

    IoConfigs.clear();
    
    Cfg.m_Port = SDI_OUTPUT_PORT;
    Cfg.m_Group = DTAPI_IOCONFIG_IOSTD;
    Cfg.m_Value = IoStdValue;
    Cfg.m_SubValue = IoStdSubValue;
    Cfg.m_ParXtra[0] = Cfg.m_ParXtra[1] = -1;
    IoConfigs.push_back(Cfg);
    dr = TheCard.SetIoConfig(IoConfigs.data(), (int)IoConfigs.size());
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed to apply IO-STD for ports");

    //.-.-.-.-.-.-.-.-.-.-.-.- Configure the internal GENREF port -.-.-.-.-.-.-.-.-.-.-.-.
    // NOTE: setup the internal GENREF port to generate a frame timing that matches the 
    // output video standard, so that we ensure all ports are aligned

    IoConfigs.clear();
    DtHwFuncDesc*  pHwf = TheCard.m_pHwf;
    for (int  i=0; i<TheCard.m_pHwf->m_DvcDesc.m_NumPorts; i++, pHwf++)
    {
        // Check if the port has GENNREF capability
        if ((pHwf->m_Flags & DTAPI_CAP_GENREF) == 0)
            continue;   // No GENREF cap

        Cfg.m_Port = pHwf->m_Port;
        Cfg.m_Group = DTAPI_IOCONFIG_GENREF;

        // Check if this is the internal GENREF (has DTAPI_CAP_VIRTUAL)?
        if ((pHwf->m_Flags & DTAPI_CAP_VIRTUAL) != 0)
            Cfg.m_Value = DTAPI_IOCONFIG_TRUE;   // The internal => enable as GENREF
        else
            Cfg.m_Value = DTAPI_IOCONFIG_FALSE;  // Not the internal => disable as GENREF
        Cfg.m_SubValue = -1;
        Cfg.m_ParXtra[0] = Cfg.m_ParXtra[1] = -1;
        // Add to list
        IoConfigs.push_back(Cfg);

        // Set the IO-standard for 'THE' GENREF port
        if (Cfg.m_Value != DTAPI_IOCONFIG_TRUE)
            continue;
        Cfg.m_Group = DTAPI_IOCONFIG_IOSTD;
        // Set a standard with matching frame rate
        DtVidStdInfo  VidInfo;
        dr = ::DtapiGetVidStdInfo(VIDSTD, LINKSTD, VidInfo);
        Cfg.m_Value=DTAPI_IOCONFIG_HDSDI; 
        int  Fps = (int)VidInfo.m_Fps;
        switch (VIDSTD)
        {
        case DTAPI_VIDSTD_2160P50B:    Fps = 25; break;
        case DTAPI_VIDSTD_2160P59_94B: Fps = 29; break;
        case DTAPI_VIDSTD_2160P60B:    Fps = 30; break;
        }
        switch (Fps)
        {
        case 23: Cfg.m_SubValue = DTAPI_IOCONFIG_720P23_98; break;
        case 24: Cfg.m_SubValue = DTAPI_IOCONFIG_720P24; break;
        case 25: Cfg.m_SubValue = DTAPI_IOCONFIG_720P25; break;
        case 29: Cfg.m_SubValue = DTAPI_IOCONFIG_720P29_97; break;
        case 30: Cfg.m_SubValue = DTAPI_IOCONFIG_720P30; break;
        case 50: Cfg.m_SubValue = DTAPI_IOCONFIG_720P50; break;
        case 59: Cfg.m_SubValue = DTAPI_IOCONFIG_720P59_94; break;
        case 60: Cfg.m_SubValue = DTAPI_IOCONFIG_720P60; break;
        }
        Cfg.m_ParXtra[0] = Cfg.m_ParXtra[1] = -1;
        // Add to list
        IoConfigs.push_back(Cfg);
    }
    dr = TheCard.SetIoConfig(IoConfigs.data(), (int)IoConfigs.size());
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed to apply GENREF IO-config");

    return (dr == DTAPI_OK);
}

//.-.-.-.-.-.-.-.-.-.-.-.-.- MxHdmi192KHzAudioDemo::OnNewFrame -.-.-.-.-.-.-.-.-.-.-.-.-.-
//
void  MxHdmi192KHzAudioDemo::OnNewFrame(DtMxData* pData, void* pOpaque)
{
    const CallbackContext*  pContext = ((const CallbackContext*)pOpaque);
    if (pContext->m_Type == CallbackContext::GEN_VIDEO)
        pContext->m_pPlayer->OnGenerateVideo(pData);
    else if (pContext->m_Type == CallbackContext::GEN_AUDIO)
        pContext->m_pPlayer->OnGenerateAudio(pData);
}

//.-.-.-.-.-.-.-.-.-.-.-.- MxHdmi192KHzAudioDemo::OnGenerateVideo -.-.-.-.-.-.-.-.-.-.-.-.
//
void  MxHdmi192KHzAudioDemo::OnGenerateVideo(DtMxData* pData)
{
    DtMxRowData&  OurRow = pData->m_Rows[OUT_ROW];

    // Must have a valid frame
    if (OurRow.m_CurFrame->m_Status != DT_FRMSTATUS_OK)
    {
        return; // Frame is not valid so we have to assume the frame buffers are unusable
    }
    // Check if any frames were skipped. Indication that we may have a performance issue
    if (pData->m_NumSkippedFrames>0)
        printf("[%lld] #skipped frames=%d\n", pData->m_Frame, pData->m_NumSkippedFrames);

    //.-.-.-.-.-.-.-.-.-.-.-.-.-.- Generate the three planes -.-.-.-.-.-.-.-.-.-.-.-.-.-.-
    // We will generate the Dutch flag in the background and overlay a moving orange 
    // rectangle

    //.-.-.-.-.-.-.-.- Step 1 : generate red, white and blue back ground -.-.-.-.-.-.-.-.-

    DtMxVideoBuf&  VideoBuf = OurRow.m_CurFrame->m_Video[0];
    // Get luminance plane
    DtMxVideoPlaneBuf&  PlaneY = VideoBuf.m_Planes[0];
    // Get chrominance plane (Cb=U)
    DtMxVideoPlaneBuf&  PlaneU = VideoBuf.m_Planes[1];
    // Get chrominance plane (Cr=V)
    DtMxVideoPlaneBuf&  PlaneV = VideoBuf.m_Planes[2];

    unsigned char* pY = PlaneY.m_pBuf;
    unsigned char* pU = PlaneU.m_pBuf;
    unsigned char* pV = PlaneV.m_pBuf;
    Color  Col(0,0,0);        
    for (int i = 0; i < VideoBuf.m_Height; i++)
    {
        
        // First third is the red band
        if (i < (VideoBuf.m_Height / 3))
        {
            Col = COL_RED;
        }
        // Second third is the white band
        else if (i < ((VideoBuf.m_Height / 3) * 2))
        {
            Col = COL_WHITE;
        }
        // Third third is the blue band
        else
        {
            Col = COL_BLUE;
        }

        // Set luminance plane
        memset(pY, Col.m_Y, VideoBuf.m_Width);
        // Set chrominance plane U
        memset(pU, Col.m_U, VideoBuf.m_Width / 2);   // Only half the U-samples
        // Set chrominance plane V
        memset(pV, Col.m_V, VideoBuf.m_Width / 2);   // Other half is V-samples

        // Next line
        pY += PlaneY.m_Stride;
        pU += PlaneU.m_Stride;
        pV += PlaneV.m_Stride;
    }
    
    //.-.-.-.-.-.-.-.-.-.-.-.-.- Step 2: overlay the rectangles -.-.-.-.-.-.-.-.-.-.-.-.-.
    for (int i = 0; i<(int)m_Rectangles.size(); i++)
    {
        MxRectangle&  R = m_Rectangles[i];

        // Move rectangle to new position
        R.Offset(R.m_Vector);

        // Bounce of the left and right edge
        if (R.Left() < m_BoundingRect.Left())
        {
            R.Offset(-R.Left() * 2, 0);
            // Invert horizontal direction of movement vector
            R.m_Vector.m_X = -R.m_Vector.m_X;
        }
        else if (R.Right() > m_BoundingRect.Right())
        {
            R.Offset((m_BoundingRect.Right() - R.Right()) * 2, 0);
            // Invert horizontal direction of movement vector
            R.m_Vector.m_X = -R.m_Vector.m_X;
        }

        // Bounce of the top and bottom edge
        if (R.Top() < m_BoundingRect.Top())
        {
            R.Offset(0, -R.Top() * 2);
            // Invert vertical direction of movement vector
            R.m_Vector.m_Y = -R.m_Vector.m_Y;
        }
        else if (R.Bottom() > m_BoundingRect.Bottom())
        {
            R.Offset(0, (m_BoundingRect.Bottom() - R.Bottom()) * 2);
            // Invert vertical direction of movement vector
            R.m_Vector.m_Y = -R.m_Vector.m_Y;
        }

        pY = PlaneY.m_pBuf + (PlaneY.m_Stride * R.Top());
        pU = PlaneU.m_pBuf + (PlaneU.m_Stride * R.Top());
        pV = PlaneV.m_pBuf + (PlaneV.m_Stride * R.Top());
        for (int j = 0; j<R.Height(); j++)
        {
            memset(pY + R.Left(), R.m_Color.m_Y, R.Width());
            memset(pU + R.Left() / 2, R.m_Color.m_U, R.Width() / 2);
            memset(pV + R.Left() / 2, R.m_Color.m_V, R.Width() / 2);

            // Next line
            pY += PlaneY.m_Stride;
            pU += PlaneU.m_Stride;
            pV += PlaneV.m_Stride;
        }
    }
}

//.-.-.-.-.-.-.-.-.-.-.-.- MxHdmi192KHzAudioDemo::OnGenerateAudio -.-.-.-.-.-.-.-.-.-.-.-.
//
void  MxHdmi192KHzAudioDemo::OnGenerateAudio(DtMxData* pData)
{
    DtMxRowData&  OurRow = pData->m_Rows[OUT_ROW];

    // Must have a valid frame
    if (OurRow.m_CurFrame->m_Status != DT_FRMSTATUS_OK)
    {
        return; // Frame is not valid so we have to assume the frame buffers are unusable
    }

    // Get frame audio data
    DtMxAudioData&  AudioData = OurRow.m_CurFrame->m_Audio;

    //.-.-.-.-.-.-.-.-.-.-.-.-.- Create the service description -.-.-.-.-.-.-.-.-.-.-.-.-.

    // 2 services = 2x 48khz stereo channel = 4x48khz channels
    // 4 services = 2x 192khz channel = 8x48khz channels
    int  NumServices = (NUM_48KHZ_AUDIO_CHANNELS /2) + (NUM_192KHZ_AUDIO_CHANNELS * 2);
    for (int  i=0; i<(int)NumServices; i++)
    {
        DtMxAudioService&  Service = AudioData.m_Services[i];
        Service.m_Valid = true;
        Service.m_SampleRate = 48000;      // 48Khz
        Service.m_ContainsData = false;    // We will insert PCM not data
        Service.m_PcmNumBits = 24;         // 24-bit audio samples
        Service.m_ServiceType = DT_AUDIOSERVICE_STEREO;    // Create a stereo pair
        Service.m_Channels.push_back(0+(i*2));   // Channel x
        Service.m_Channels.push_back(1+(i*2));   // Channel x+1
        AudioData.m_Channels[0+i*2].m_Service = i;
        AudioData.m_Channels[1+i*2].m_Service = i;
    }

    //-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Add PCM samples -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
    
    // Assume 4 channels of 48khz and 2 channels of 192khz
    static MxToneGeneratorEx  TONE_GEN[] = 
    {
        // 2.0KHz tone; with random volume
        MxToneGeneratorEx(AUDIO_MODE, 1, 2000, 0.4 / ((rand()%4)+1), 48000),
        // 2.5KHz tone; with random volume
        MxToneGeneratorEx(AUDIO_MODE, 2, 2500, 0.4 / ((rand()%4)+1), 48000),
        // 3.0KHz tone; with random volume
        MxToneGeneratorEx(AUDIO_MODE, 3, 3000, 0.4 / ((rand()%4)+1), 48000),
        // 3.5KHz tone; with random volume
        MxToneGeneratorEx(AUDIO_MODE, 4, 3500, 0.4 / ((rand()%4)+1), 48000),
        // 1.0KHz tone; with random volume
        MxToneGeneratorEx(AUDIO_MODE, -1, 1000, 0.4 / ((rand()%4)+1), 192000),
        // 1.5KHz tone; with random volume
        MxToneGeneratorEx(AUDIO_MODE, -1, 1500, 0.4 / ((rand()%4)+1), 192000),
    };

    // Generate 48kHz audio
    int  ChannelOffset = 0;
    for (int  i=0; i<(int)NUM_48KHZ_AUDIO_CHANNELS; i++) // 4x 48khz channels
    {
        DtMxAudioChannel&  Channel = AudioData.m_Channels[ChannelOffset+i];
            
        // Get number of samples for 1 48khz channel
        unsigned int  NumSamples = Channel.m_NumSamplesHint;
        
        // Generate 48kHz audio
        TONE_GEN[i].Generate(Channel.m_pBuf, NumSamples);
        Channel.m_NumValidSamples = NumSamples;
        if (AUDIO_MODE == DT_AUDIO_SAMPLE_PCM)
        {
            // Init the channel status word with 48khz
            InitChannelStatus(Channel, 48000);
        }
    }

    // Generate 192kHz audio
    ChannelOffset = NUM_48KHZ_AUDIO_CHANNELS;
    for (int  i=0; i<(int)NUM_192KHZ_AUDIO_CHANNELS; i++)   // 2x 192KHz channels
    {
        // Get first channel of group of 4 and calculate 
        // number of 192KHz samples to insert
        int  NumSamples = AudioData.m_Channels[ChannelOffset+i*4].m_NumSamplesHint*4;
        
        // Create buffer for 192khz samples
        if (m_pAudioBuffer[i] == NULL)
        {
            m_pAudioBuffer[i] = new unsigned int[NumSamples];
        } else {
            if (m_AudioBufferLength[i] < (unsigned int)NumSamples)
            {
                delete [] m_pAudioBuffer[i];
                m_pAudioBuffer[i] = new unsigned int [NumSamples];
            }
        }
        m_AudioBufferLength[i] = NumSamples;

        // Generate 192kHz audio
        TONE_GEN[ChannelOffset+i].Generate(m_pAudioBuffer[i], NumSamples);

        // Interleave the 192khz audio samples into the 4 channels
        for (int  j=0; j<4; j++)
        {
            int  Index = j + (ChannelOffset+i*4);
            DtMxAudioChannel&  Channel = AudioData.m_Channels[Index];
            for (int k=0; k<NumSamples/4;k++)
            {
                Channel.m_pBuf[k] = m_pAudioBuffer[i][(k*4)+j];
            }
            Channel.m_NumValidSamples = NumSamples / 4;
            if (AUDIO_MODE == DT_AUDIO_SAMPLE_PCM)
            {
                // Init the channel status word with 192khz
                InitChannelStatus(Channel, 192000);
            }
        }
    }
}

//-.-.-.-.-.-.-.-.-.-.-.- MxHdmi192KHzAudioDemo::InitChannelStatus -.-.-.-.-.-.-.-.-.-.-.-
//
void MxHdmi192KHzAudioDemo::InitChannelStatus(DtMxAudioChannel& Channel, int AudioFs)
{
    MxAudioStatusWordParser  Parser;

    // Init all status word for all channels in the service
    Channel.m_Status.m_Valid = false;

    // Clear to all zero's
    memset(Channel.m_Status.m_Data, 0x00, sizeof(Channel.m_Status.m_Data));

    // Attach status word parser
    Parser.Attach(Channel.m_Status.m_Data);

    // Is professional by definition
    Parser.SetProfessionalUse(MxAudioStatusWordParser::MX_USE_PROFESSIONAL);

    // Set audio sample frequency
    Parser.SetEncodedSamplingFreqHz(AudioFs);

    // Set PCM or data
    Channel.m_Status.SetPcmAudio(true);
    // Set number of bits
    Channel.m_Status.SetPcmNumBits(24, 0);

    // Set the channel modes
    Parser.SetChannelMode(MxAudioStatusWordParser::MX_CHANMODE_STEREO);
    
    // Update the CRC8 
    Parser.UpdateCrc8();

    // Mark status word as valid
    Channel.m_Status.m_Valid = true;
}

//=+=+=+=+=+=+=+=+=+=+=+=+ MxAudioStatusWordParser implementation +=+=+=+=+=+=+=+=+=+=+=+=

// Init static members
const unsigned char MxAudioStatusWordParser::BIT_MASK[9] =
{
    0x00, 0x01, 0x03, 0x07, 0x0F, 0x01F, 0x3f, 0x7F, 0xFF
};


//-.-.-.-.-.-.-.-.-.- MxAudioStatusWordParser::MxAudioStatusWordParser -.-.-.-.-.-.-.-.-.-
//
MxAudioStatusWordParser::MxAudioStatusWordParser() 
    : m_pData(NULL)
{
     // Init CRC8 table.
    // Polynomial = X^8+X^4+X^3+X^2+1 => 0x1D
    for (int  i=0; i<256; i++)
    {
        unsigned char  Crc = i;
        for (int  j=0; j<8; j++)
            Crc = (Crc<<1) ^ ((Crc&0x80)!=0 ? 0x1D : 0);
        m_Crc8Table[i] = Crc;
    }
}

//.-.-.-.-.-.-.-.-.- MxAudioStatusWordParser::~MxAudioStatusWordParser -.-.-.-.-.-.-.-.-.-
//
MxAudioStatusWordParser::~MxAudioStatusWordParser()
{
}

//-.-.-.-.-.-.-.-.-.-.-.-.- MxAudioStatusWordParser::ComputeCrc8 -.-.-.-.-.-.-.-.-.-.-.-.-
//
unsigned char  MxAudioStatusWordParser::ComputeCrc8()
{
    MX_ASSERT(m_pData != NULL);

    // Initial value should be all ones
    unsigned char  Crc = 0xFF;
    for (int  i=0; i<23; i++)
        Crc = m_Crc8Table[Crc ^ m_pData[i]];
    return Crc;
}

//.-.-.-.-.-.-.-.-.- MxAudioStatusWordParser::GetEncodedSamplingFreqHz -.-.-.-.-.-.-.-.-.-
//
int  MxAudioStatusWordParser::GetEncodedSamplingFreqHz()
{
    switch (GetEncodedSamplingFreq())
    {
    case MX_ENCSAMPLFREQ_32000: return 32000;
    case MX_ENCSAMPLFREQ_44100: return 44100;
    case MX_ENCSAMPLFREQ_48000: return 48000;
    }
    return GetSamplingFreqHz();
}
// Setter
void  MxAudioStatusWordParser::SetEncodedSamplingFreqHz(int  FreqHz)
{
    switch (FreqHz)
    {
    case 32000: SetEncodedSamplingFreq(MX_ENCSAMPLFREQ_32000); break;
    case 44100: SetEncodedSamplingFreq(MX_ENCSAMPLFREQ_44100); break;
    case 48000: SetEncodedSamplingFreq(MX_ENCSAMPLFREQ_48000); break;
    default:    
        SetEncodedSamplingFreq(MX_ENCSAMPLFREQ_NOTSET); 
        SetSamplingFreqHz(FreqHz);
        break;
    }
}

//.-.-.-.-.-.-.-.-.-.-.- MxAudioStatusWordParser::GetSamplingFreqHz -.-.-.-.-.-.-.-.-.-.-.
//
int  MxAudioStatusWordParser::GetSamplingFreqHz()
{
    MX_ASSERT(m_pData != NULL);
    switch (GetSamplingFrequency())
    {
    case MX_SAMPLFREQ_NOTSET:   return 48000;
    case MX_SAMPLFREQ_22050:    return 22050;
    case MX_SAMPLFREQ_24000:    return 24000;
    case MX_SAMPLFREQ_88200:    return 88200;
    case MX_SAMPLFREQ_96000:    return 96000;
    case MX_SAMPLFREQ_176400:   return 176400;
    case MX_SAMPLFREQ_192000:   return 192000;
    case MX_SAMPLFREQ_USRDEF:   return -1;
    default:
        MX_ASSERT(false);
        break;
    }
    return 48000;
}
// SETTER
void  MxAudioStatusWordParser::SetSamplingFreqHz(int  FreqHz)
{
    MX_ASSERT(m_pData != NULL);
    switch (FreqHz)
    {
    case 22050:     SetSamplingFrequency(MX_SAMPLFREQ_22050); break;
    case 24000:     SetSamplingFrequency(MX_SAMPLFREQ_24000); break;
    case 88200:     SetSamplingFrequency(MX_SAMPLFREQ_88200); break;
    case 96000:     SetSamplingFrequency(MX_SAMPLFREQ_96000); break;
    case 176400:    SetSamplingFrequency(MX_SAMPLFREQ_176400); break;
    case 192000:    SetSamplingFrequency(MX_SAMPLFREQ_192000); break; 
    case -1:        SetSamplingFrequency(MX_SAMPLFREQ_USRDEF); break;
    default:        SetSamplingFrequency(MX_SAMPLFREQ_NOTSET); break;
    }
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.- MxAudioStatusWordParser::Attach -.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
void  MxAudioStatusWordParser::Attach(unsigned char  Data[24])
{
    MX_ASSERT(Data != NULL);
    m_pData = Data;
}

//+=+=+=+=+=+=+=+=+=+=+=+=+=+ MxToneGeneratorEx implementation +=+=+=+=+=+=+=+=+=+=+=+=+=+

// Parity lookup table
const bool  MxToneGeneratorEx::PARITY_TABLE256[256] = 
{
    #define P2(n) n, n^1, n^1, n
    #define P4(n) P2(n), P2(n^1), P2(n^1), P2(n)
    #define P6(n) P4(n), P4(n^1), P4(n^1), P4(n)
            P6(0), P6(1), P6(1), P6(0)
};
#undef P2
#undef P4
#undef P6


//-.-.-.-.-.-.-.-.-.-.-.-.- MxToneGeneratorEx::MxToneGeneratorEx -.-.-.-.-.-.-.-.-.-.-.-.-
//
MxToneGeneratorEx::MxToneGeneratorEx(const DtMxAudioSampleType AudioType, int  Channel, 
                                                      int  ToneFreq, double  Attn, int Fs)
    : MxToneGenerator(ToneFreq, Attn, Fs),
      m_Channel(Channel),
      m_AudioSampleFormat(AudioType)
{
    for (int i=0; i< MAX_NUM_CHANNELS; i++)
    {
        m_StatusWordIsValid[i] = false;
        m_NextStatusWordBit[i] = 0;
    }
    if (m_AudioSampleFormat == DT_AUDIO_SAMPLE_AES3)
        Init();
}


//.-.-.-.-.-.-.-.-.-.-.-.- MxToneGeneratorEx::~MxToneGeneratorEx -.-.-.-.-.-.-.-.-.-.-.-.-
//
MxToneGeneratorEx::~MxToneGeneratorEx()
{
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- MxToneGeneratorEx::Init -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
void MxToneGeneratorEx::Init()
{
    // Clear to all zero's
    for (int i=0; i<MAX_NUM_CHANNELS; i++)
    {
        memset(m_StatusWord[i], 0x00, sizeof(m_StatusWord[i]));

        MxAudioStatusWordParser  Parser;
        Parser.Attach(m_StatusWord[i]);

        // Is professional by definition
        Parser.SetProfessionalUse(MxAudioStatusWordParser::MX_USE_PROFESSIONAL);
        Parser.SetEncodedSamplingFreqHz(m_Fs);

        // Set PCM or data
        Parser.SetLinearPcm(MxAudioStatusWordParser::MX_LINEARPCM_PCM);
        // Set number of bits
    
        // FOR NOW: assume 24-bit
        Parser.SetUseOfAuxSampleBits(MxAudioStatusWordParser::MX_AUXBITS_24_BIT);
        Parser.SetAudioSampleWordLength(MxAudioStatusWordParser::MX_SAMPLELENGTH_MAX);
    
        // Set the channel modes
        Parser.SetChannelMode(MxAudioStatusWordParser::MX_CHANMODE_STEREO);
    
        // Update the CRC8 
        Parser.UpdateCrc8();

        m_NextStatusWordBit[i] = 0;
        m_StatusWordIsValid[i] = true;
    }
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.- MxToneGeneratorEx::GenerateAes3 -.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
void  MxToneGeneratorEx::Generate(unsigned int* pSamples,int NumSamples2Gen)
{
    switch (m_AudioSampleFormat)
    {
    case DT_AUDIO_SAMPLE_PCM:  MxToneGenerator::Generate(pSamples, NumSamples2Gen); break;
    case DT_AUDIO_SAMPLE_AES3: GenerateAes3(pSamples, NumSamples2Gen); break;
    default:                   MX_ASSERT(1 == 0); break;
    }
}
// OVERLOAD: generates AES3 samples
void  MxToneGeneratorEx::GenerateAes3(unsigned int*  pSamples,int  NumSamples2Gen)
{
    
    //-.-.-.-.-.-.-.-.-.-.-.-.-.- Step 1: generate PCM samples -.-.-.-.-.-.-.-.-.-.-.-.-.-

    // Create buffer for PCM samples
    std::vector<unsigned int> PcmSamples;
    PcmSamples.resize(NumSamples2Gen);

    // Generate the samples
    MxToneGenerator::Generate(PcmSamples.data(), (int)PcmSamples.size());

    //-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Step 2: convert to AES3 -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.

    MxAudioStatusWordParser  Parser[MAX_NUM_CHANNELS];
    for (int i=0; i< MAX_NUM_CHANNELS; i++)
        Parser[i].Attach(m_StatusWord[i]);

    if (m_Fs != 192000)
    {
        // Generate AES bytes for 1 channel
        for(int i=0; i<(int)PcmSamples.size(); i++)
        {
            pSamples[i] = (PcmSamples[i]>>4) & 0x0FFFFFF0;

            // Set Z-bit is first bit in sequence, but only set on odd-channels
            const int  ChannelIndex = (m_Channel-1) % 4;
            unsigned char  ZBit = (m_NextStatusWordBit[0]==0) ? 4 : 0;
            MX_ASSERT((ZBit==0 || ZBit==4) && (ChannelIndex>=0 && ChannelIndex<=3));
            static const unsigned int Z_BIT[8] = 
            { 
                // Z=0:I=0   Z=0:I=1      Z=0:I=2      Z=0:I=3
                0x00000000U, 0x00000000U, 0x00000000U, 0x00000000U, 
                // Z=1:I=0   Z=1:I=1      Z=1:I=2      Z=1:I=3
                0x00000008U, 0x00000000U, 0x00000008U, 0x00000000U, 
            };
            pSamples[i] |= Z_BIT[ZBit | ChannelIndex];
                
            // Set C-Bit
            unsigned char  CBit = Parser[0].GetBits(m_NextStatusWordBit[0], 1);
            MX_ASSERT(CBit>=0 && CBit<=1);
            static const unsigned int C_BIT[2] = { 0x00000000U, 0x40000000U };
            pSamples[i] |= C_BIT[CBit];
                
            // Set P-bit (parity)
            static const unsigned int  P_BIT[2] = { 0x00000000U, 0x80000000U };
            pSamples[i] |= P_BIT[Parity32(pSamples[i])];
            m_NextStatusWordBit[0]++; m_NextStatusWordBit[0] %= 192;
        }
    } else {
        // Generate AES3 bytes for 4 channels (192kHz)
        int ChannelIndex = 0;
        for(int i=0; i<(int)PcmSamples.size(); i++)
        {
            pSamples[i] = (PcmSamples[i]>>4) & 0x0FFFFFF0;
            
            // Set Z-bit is first bit in sequence, but only set on odd-channels
            unsigned char  ZBit = (m_NextStatusWordBit[ChannelIndex]==0) ? 4 : 0;
            MX_ASSERT((ZBit==0 || ZBit==4) && (ChannelIndex>=0 && ChannelIndex<=3));
            static const unsigned int Z_BIT[8] = 
            { 
                // Z=0:I=0   Z=0:I=1      Z=0:I=2      Z=0:I=3
                0x00000000U, 0x00000000U, 0x00000000U, 0x00000000U, 
                // Z=1:I=0   Z=1:I=1      Z=1:I=2      Z=1:I=3
                0x00000008U, 0x00000000U, 0x00000008U, 0x00000000U, 
            };
            pSamples[i] |= Z_BIT[ZBit | ChannelIndex];
                
            // Set C-Bit
            unsigned char  CBit = 
                       Parser[ChannelIndex].GetBits(m_NextStatusWordBit[ChannelIndex], 1);
            MX_ASSERT(CBit>=0 && CBit<=1);
            static const unsigned int C_BIT[2] = { 0x00000000U, 0x40000000U };
            pSamples[i] |= C_BIT[CBit];
                
            // Set P-bit (parity)
            static const unsigned int  P_BIT[2] = { 0x00000000U, 0x80000000U };
            pSamples[i] |= P_BIT[Parity32(pSamples[i])];

            m_NextStatusWordBit[ChannelIndex]++; m_NextStatusWordBit[ChannelIndex] %= 192;
            ChannelIndex++;
            ChannelIndex = ChannelIndex % 4;
        }
    }
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.- MxToneGeneratorEx::Parity32 -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
// Compute parity of 32-bit word
//
bool  MxToneGeneratorEx::Parity32(unsigned int  Data)
{
#ifdef _WIN32
    Data ^= Data >> 16;
    Data ^= Data >> 8;
    return PARITY_TABLE256[Data & 0xff];
#else
    return  __builtin_parity(Data);
#endif
}

} // namespace MxExamples
