//*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#* DtMxPipDemo.cpp *#*#*#*#*#*#*#* (C) 2014-2015 DekTec
//
// This example application demonstrates a matrix configuration which has one input-output
// row and one input-only row. The video from the input-only row is scaled and is
// overlayed as a PiP (Picture-in-Picture) on the video of input-output row.
//
// NOTE:
//   This application assumes a valid signal is connected to ports 1&3 of your DTA-2174
//   and it generates an output on port 2 (see MxPipDemo::Init)
//

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Include files -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
#include <cstdio>
#include <cstring>
#ifdef _WIN32
    // Windows specific includes
    #include <windows.h>
    #include <conio.h>
#else
    // Linux specific includes
    #include <unistd.h>
    #include <limits.h>
    #include <termios.h>
    #include <sys/select.h>
#endif
#include "MxPipDemo.h"

// All code resides in MxExamples namespace
using namespace MxExamples;

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Application Version -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
#define APP_NAME                "DtMxPipDemo"
#define APP_VERSION_MAJOR       1
#define APP_VERSION_MINOR       1
#define APP_VERSION_BUGFIX      0

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Application configuration -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
// Card to attach to
static const int  CARD_TYPE = 2174;
// Number of card to attach to
static const int  CARD_NO = 0;
// Input and output ports
static const int  IN_PORT=1, OUT_PORT=2, PIP_PORT=3;
// Row indices
static const int  INOUT_ROW=0, PIP_ROW = 1;

// Enable this define to get profiling information
//#define  MX_PRINT_PROFILING_INFO

//=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+ Main loop +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+



//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- main -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
int main(int  argc, char*  argv[])
{
    DTAPI_RESULT  dr;
    DtDevice  TheCard;
    MxPipDemo  PipDemo;

#ifndef _WIN32
    ChangeTerminalMode(1);  // Set terminal mode, so that we can use _kbhit()
#endif

    try
    {
        //-.-.-.-.-.-.-.-.-.-.-.-.-.- Display welcome message -.-.-.-.-.-.-.-.-.-.-.-.-.-.

        printf("%s V%d.%d.%d (c) 2014-2015 DekTec Digital Video B.V.\n\n", APP_NAME,
                                APP_VERSION_MAJOR, APP_VERSION_MINOR, APP_VERSION_BUGFIX);

        //.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Attach to the card -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.

        dr = TheCard.AttachToType(CARD_TYPE, CARD_NO);
        if (dr != DTAPI_OK)
            MX_THROW_EXC(dr, "Failed to attach to DTA-%d", CARD_TYPE);

        // Check the ports for CAP_MATRIX2. The DtMxXxx API is only supported by ports 
        // with DTAPI_CAP_MATRIX2 capability
        if ((TheCard.m_pHwf[(IN_PORT-1)].m_Flags & DTAPI_CAP_MATRIX2) == 0)
            MX_THROW_EXC_NO_ERR("Port %d does not have the MATRIX2 capability", IN_PORT);
        if ((TheCard.m_pHwf[(OUT_PORT-1)].m_Flags & DTAPI_CAP_MATRIX2) == 0)
            MX_THROW_EXC_NO_ERR("Port %d does not have the MATRIX2 capability", OUT_PORT);
        if ((TheCard.m_pHwf[(PIP_PORT-1)].m_Flags & DTAPI_CAP_MATRIX2) == 0)
            MX_THROW_EXC_NO_ERR("Port %d does not have the MATRIX2 capability", PIP_PORT);
                
        //.-.-.-.-.-.-.-.-.-.-.-.-.-.- Create the PIP matrix -.-.-.-.-.-.-.-.-.-.-.-.-.-.-

        // Init the matrix
        if (!PipDemo.Init(TheCard))
            MX_THROW_EXC_NO_ERR("Failed to init demo");
        // Start the matrix
        if (!PipDemo.Start())
            MX_THROW_EXC_NO_ERR("Failed to start demo");

        //-.-.-.-.-.-.-.-.-.- Process loop, wait for user to press key -.-.-.-.-.-.-.-.-.-
        
        printf("\nPress any key to exit\n\n");
        while(!_kbhit()) { Sleep(10); }
        _getch();   // Read the keystroke from the console

        PipDemo.Stop(); // Stop the matrix

        // Print profling information
#ifdef MX_PRINT_PROFILING_INFO
        PipDemo.PrintProfilingInfo();
        printf("\nPress any key to exit\n");
        // Wait for a keystroke before exiting. To allow user time to take in the 
        // profiling information
        _getch();
#endif
    }
    catch(MxDemoExc  e)
    {
        puts(e.ErrorStr());
    }
    catch(...) 
    {
        puts("Encountered an unhandled exception!!!");
    }

#ifndef _WIN32
    ChangeTerminalMode(0);  // Restore terminal mode
#endif
    return 0;
}

// All code resides in MxExamples namespace
namespace MxExamples
{

//+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+ MxPipDemo implementation +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- MxPipDemo::MxPipDemo -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
MxPipDemo::MxPipDemo()
{
}

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- MxPipDemo::~MxPipDemo -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
MxPipDemo::~MxPipDemo()
{
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- MxPipDemo::Init -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
bool  MxPipDemo::Init(DtDevice&  TheCard)
{
    DTAPI_RESULT  dr;

    //-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Prep card for matrix -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-

    // The card must have been attached
    if (!TheCard.IsAttached())
        return false;
    if (!PrepCard(TheCard))
        return false;

    //.-.-.-.-.-.-.-.-.-.-.-.-.- Configure the input-ouput row -.-.-.-.-.-.-.-.-.-.-.-.-.-

    // Attach row to the input
    DtMxPort  InPort;
    dr = InPort.AddPhysicalPort(&TheCard, IN_PORT);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed to add port %d as physical in-port", IN_PORT);
    dr = m_TheMatrix.AttachRowToInput(INOUT_ROW, InPort);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed to attach row %d to in-port %d", INOUT_ROW, IN_PORT);
    
    // Attach row to the output
    DtMxPort  OutPort;
    dr = OutPort.AddPhysicalPort(&TheCard, OUT_PORT);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed to add port %d as physical out-port", OUT_PORT);
    dr = m_TheMatrix.AttachRowToOutput(INOUT_ROW, OutPort);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed to attach row %d to out-port %d", INOUT_ROW, OUT_PORT);
    
    DtMxRowConfig  Row1Config;
    Row1Config.m_Enable = true;
    Row1Config.m_RowSize = 1;       // Keep a history of one frame

    // Enable the video and get full picture
    Row1Config.m_VideoEnable = true;
    Row1Config.m_Video.m_PixelFormat = DT_PXFMT_UYVY422_8B; // Native packed format in 8B
    Row1Config.m_Video.m_StartLine1 = Row1Config.m_Video.m_StartLine2 = 1;
    Row1Config.m_Video.m_NumLines1 = Row1Config.m_Video.m_NumLines2 = -1;
    Row1Config.m_Video.m_Scaling = DTAPI_SCALING_OFF;   // No scaling for this row
    Row1Config.m_Video.m_LineAlignment = 8; // Want an 8-byte alignment

    // Apply the row configuration
    dr = m_TheMatrix.SetRowConfig(INOUT_ROW, Row1Config);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed set row %d config", INOUT_ROW);
       
    //-.-.-.-.-.-.-.-.-.-.-.-.-.- Configure the input-only row -.-.-.-.-.-.-.-.-.-.-.-.-.-

    // Attach row to the input
    DtMxPort  PipPort;
    dr = PipPort.AddPhysicalPort(&TheCard, PIP_PORT);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed to add port %d as physical in-port", PIP_PORT);
    dr = m_TheMatrix.AttachRowToInput(PIP_ROW, PipPort);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed to attach row %d to in-port %d", PIP_ROW, PIP_PORT);

    DtMxRowConfig  Row2Config;
    Row2Config.m_Enable = true;
    Row2Config.m_RowSize = 1;       // Keep a history of one frame

    // Enable the video and get scaled 1/16 picture
    Row2Config.m_VideoEnable = true;
    Row2Config.m_Video.m_PixelFormat = DT_PXFMT_UYVY422_8B; // Native packed format in 8B
    Row2Config.m_Video.m_StartLine1 = Row2Config.m_Video.m_StartLine2 = 1;
    Row2Config.m_Video.m_NumLines1 = Row2Config.m_Video.m_NumLines2 = -1;
    Row2Config.m_Video.m_Scaling = DTAPI_SCALING_1_16;   // 1/16th scaling for this row
    Row2Config.m_Video.m_LineAlignment = 8; // Want an 8-byte alignment

    // Apply the row configuration
    dr = m_TheMatrix.SetRowConfig(PIP_ROW, Row2Config);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed set row %d config", PIP_ROW);

    // Use a single phase
    dr = m_TheMatrix.SetNumPhases(1);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed set number of phases");

    //-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- Final configuration -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.

    // Determine the default recommended end-to-end delay for this configuration
    int  DefEnd2EndDelay = 0;
    double CbFrames=0.0;    // Time available for the call-back method
    dr = m_TheMatrix.GetDefEndToEndDelay(DefEnd2EndDelay, CbFrames);
    printf("  Default end-to-end delay = %d Frames\n", DefEnd2EndDelay);
    printf("  Time-for-Callback = %.1f Frame Period\n", CbFrames);
    // Apply the default end-to-end delay
    dr = m_TheMatrix.SetEndToEndDelay(DefEnd2EndDelay);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed set the end-to-end delay (delay=%d)", DefEnd2EndDelay);

    // Register our callback
    dr = m_TheMatrix.AddMatrixCbFunc(MxPipDemo::OnNewFrame, this);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed register our callback function");

    return true;
}

//-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- MxPipDemo::PrepCard -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.
//
bool  MxPipDemo::PrepCard(DtDevice&  TheCard)
{
    DTAPI_RESULT  dr;
           
    //-.-.-.-.-.-.-.-.-.-.-.-.-.- Configure port io-direction -.-.-.-.-.-.-.-.-.-.-.-.-.-.

    // Configure IN_PORT as input
    dr = TheCard.SetIoConfig(IN_PORT, DTAPI_IOCONFIG_IODIR, DTAPI_IOCONFIG_INPUT, 
                                                                    DTAPI_IOCONFIG_INPUT);
     if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed to configure port %d as input", IN_PORT);

    // Configure OUT_PORT as output
    dr = TheCard.SetIoConfig(OUT_PORT, DTAPI_IOCONFIG_IODIR, DTAPI_IOCONFIG_OUTPUT, 
                                                                   DTAPI_IOCONFIG_OUTPUT);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed to configure port %d as output", OUT_PORT);
    // Configure PIP_PORT as input
    dr = TheCard.SetIoConfig(PIP_PORT, DTAPI_IOCONFIG_IODIR, DTAPI_IOCONFIG_INPUT, 
                                                                    DTAPI_IOCONFIG_INPUT);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed to configure port %d as input", PIP_PORT);

    //.-.-.-.-.- Detect applied IO-standard and configure all ports accordingly -.-.-.-.-.

    int  IoStdValue=-1, IoStdSubValue=-1;
    do
    {
        IoStdValue=-1, IoStdSubValue=-1;
        dr = TheCard.DetectIoStd(IN_PORT, IoStdValue, IoStdSubValue);
        Sleep(10);

        if (_kbhit())
        {
            _getch();   // read keystroke from console
            return false;
        }
    }
    while (dr != DTAPI_OK);

    int  VidStd = DTAPI_VIDSTD_UNKNOWN;
    ::DtapiIoStd2VidStd(IoStdValue, IoStdSubValue, VidStd);
    printf("  Detected video standard = %s\n", ::DtapiVidStd2Str(VidStd));
    // Get frame details for the detected standard
    ::DtapiGetVidStdInfo(VidStd, m_VidInfo);
    
    // Configure all three ports for with detected standard
    DtIoConfig  IoConfig[3];
    IoConfig[0].m_Port = IN_PORT;
    IoConfig[1].m_Port = OUT_PORT;
    IoConfig[2].m_Port = PIP_PORT;
    for (int  i=0; i<3; i++)
    {
        IoConfig[i].m_Group = DTAPI_IOCONFIG_IOSTD;
        IoConfig[i].m_Value = IoStdValue;
        IoConfig[i].m_SubValue = IoStdSubValue;
        IoConfig[i].m_ParXtra[0] = IoConfig[i].m_ParXtra[1] = -1;
    }
    dr = TheCard.SetIoConfig(IoConfig, 3);
    if (dr != DTAPI_OK)
        MX_THROW_EXC(dr, "Failed to apply IO-STD for ports ");

    return (dr == DTAPI_OK);
}

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- MxPipDemo::OnNewFrame -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
// Static callback. Forwards the callback to actual handler
//
void  MxPipDemo::OnNewFrame(DtMxData* pData, void* pOpaque)
{
    // Forward to actual implementation
    ((MxPipDemo*)pOpaque)->OnNewFrame(pData);
}

//.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.- MxPipDemo::OnNewFrame -.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-.-
//
void  MxPipDemo::OnNewFrame(DtMxData* pData)
{
    DtMxRowData&  InOutRow = pData->m_Rows[INOUT_ROW];
    DtMxRowData&  PipRow = pData->m_Rows[PIP_ROW];

    // Do we have an interlaced or progressive format
    int  NumFields = (m_VidInfo.m_IsInterlaced) ? 2 : 1;

    // Check if we have a valid input. If not init the video buffer with a pattern, so 
    // that we at least have a output with some video
    if (InOutRow.m_CurFrame->m_Status!=DT_FRMSTATUS_OK && 
                                    InOutRow.m_CurFrame->m_Status!=DT_FRMSTATUS_DUPLICATE)
    {
        // Init video buffer with a pattern
        for (int  i=0; i<NumFields; i++)
            InOutRow.m_CurFrame->m_Video[i].InitBuf(DT_VIDPAT_WHITE_FRAME);
    }
    // Must have a valid PIP
    if (PipRow.m_CurFrame->m_Status!=DT_FRMSTATUS_OK && 
                                      PipRow.m_CurFrame->m_Status!=DT_FRMSTATUS_DUPLICATE)
    {
        return; // Nothing to overlay
    }

    // Offset for insert position of top left corner of the PIP image
    // 1. Vertical offset in #lines. At 1/20th of the field 
    const int  VERT_OFFSET_IN_LINES = (m_VidInfo.m_NumLines / NumFields) / 20;    
    // 2. Horizontal offset in #pixels. At 1/20th of the width
    const int  HORZ_OFFSET_IN_PIXELS = InOutRow.m_CurFrame->m_Video[0].m_Width / 20;    
    
    // Overlay the PIP over the input video
    for (int  i=0; i<NumFields; i++)
    {
        // In-out video is destination buffer
        unsigned char*  pDest = InOutRow.m_CurFrame->m_Video[i].m_Planes[0].m_pBuf;
        int  DestStride = InOutRow.m_CurFrame->m_Video[i].m_Planes[0].m_Stride;

        // PIP video is our source data
        unsigned char*  pSrc = PipRow.m_CurFrame->m_Video[i].m_Planes[0].m_pBuf;
        int  SrcStride = PipRow.m_CurFrame->m_Video[i].m_Planes[0].m_Stride;
        int  SrcWidth = PipRow.m_CurFrame->m_Video[i].m_Width;

        // Apply offset 
        pDest += DestStride * VERT_OFFSET_IN_LINES;  // Vertical offset
        pDest += HORZ_OFFSET_IN_PIXELS * 2;  // Horizontal offset (1-pixel = 2-symbols)

        // Copy all PIP video lines
        for (int  j=0; j<PipRow.m_CurFrame->m_Video[i].m_Height; j++)
        {
            // Copy pixels (1-pixel = 2-symbols)
            memcpy(pDest, pSrc, SrcWidth*2);
            // Move to next line (have to move with the stride size)
            pDest += DestStride; pSrc += SrcStride;
        }
    }   
}

} // namespace MxExamples
